# -*- coding: utf-8 -*-
"""

    mslib.mswms.mss_2D_sections
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    2D cross-section style super classes for use with the
    MSSPlotDriver subclasses.

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2018 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging
from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass


class Abstract2DSectionStyle(with_metaclass(ABCMeta, object)):
    """
    Horizontal Section Styles
    Abstract 2D section super class. Use this class as a parent
    for classes implementing horizontal and vertical cross sections.
    """

    # Name and title of each derived style have to be defined if the style
    # should be used with the WMS WSGI module.
    name = None
    title = None
    abstract = None
    queryable = False
    styles = None

    # Define the datafields required by a style with this property.
    required_datafields = []

    def __init__(self, driver=None):
        self.set_driver(driver)

    def required_datatypes(self):
        """Returns a list containing the datatypes required by the
           data fields requested by the style.
        """
        result = set([datafield[0] for datafield in self.required_datafields])
        if len(result) > 2 and "sfc" not in result:
            msg = "A Plot may contain only 'sfc' and *one* 4-D type! ({}: {})".format(type(self), result)
            logging.fatal(msg)
            raise RuntimeError(msg)
        return result

    def _prepare_datafields(self):
        """Optional re-implementation: Use this function to process some
           input data before the plotting starts (e.g. to derive potential
           temperature from pressure and temperature).
        """
        pass

    def supported_epsg_codes(self):
        """Returns a list of supported EPSG codes, if available.
        """
        return []

    @abstractmethod
    def supported_crs(self):
        """Returns a list of the coordinate reference systems supported by
           this style.
        """
        pass

    def set_driver(self, driver):
        """Set the driver object for this layer.
        """
        self.driver = driver

    def get_init_times(self):
        """Returns a list of available forecast init times (base times).
        """
        if self.driver is not None:
            return self.driver.get_init_times()
        else:
            return []

    def get_all_valid_times(self):
        """Returns a list containing the combined forecast valid times of
           all available init times.
        """
        valid_times = set()
        if self.driver is not None:
            for vartype, varname in self.required_datafields:
                vtimes = self.driver.get_all_valid_times(varname, vartype)
                if len(valid_times) == 0:
                    valid_times.update(vtimes)
                else:
                    valid_times.intersection_update(vtimes)
        valid_times = list(valid_times)
        valid_times.sort()
        return valid_times

    def uses_time_dimensions(self):
        """Returns whether this layer uses the WMS time dimensions. If False,
           valid_time and init_time do not have to be specified to
           plot_hsection().

        Currently implemented by testing whether the style requires data fields
        from the ECMWF forecast.
        """
        return len(self.required_datafields) > 0

    def get_elevations(self):
        """Returns a list of available elevations for this layer.

        Assumes that the same elevation levels are available for all time
        steps.
        """
        logging.debug(u"checking vertical dimensions for layer '%s'.", self.name)
        successful = False
        if self.driver is not None and not all(_x[0] in ["sfc"] for _x in self.required_datafields):
            # Get the latest init time.
            init_times = self.driver.get_init_times()
            valid_times = self.get_all_valid_times()

            if len(init_times) == 0:
                logging.error("ERROR: cannot determine initialisation time(s) "
                              "of this dataset. Check that file structure is "
                              "in accordance with the used access class in mss_wms_settings.py!")
                return []

            # Open the data files for the analysis (valid time = init time).
            # It can happen that not all required files are available for
            # all init times. The try..except block takes care of this
            # problem and tries to find a time at which all required
            # files are available.
            for time in init_times:
                try:
                    valid_time = time
                    if valid_time not in valid_times:
                        logging.error("init_time is not contained in valid_times")
                        valid_time = valid_times[0]
                    self.driver.set_plot_parameters(
                        self, init_time=time, valid_time=valid_time)
                except (IOError, ValueError) as ex:
                    logging.debug(u"WARNING: unsuccessfully examined data for "
                                  u"init time '%s'.. trying next time.", time)
                    logging.debug(u"(Error message: %s %s)", type(ex), ex)
                else:
                    successful = True
                    break

            # If the analysis time is not available try to extract the available
            # valid times for the datafield required by this style.
            if not successful:
                vartype, varname = "", ""
                if len(self.required_datafields) > 0:
                    vartype, varname = self.required_datafields[0]
                for it in init_times:
                    valid_times = self.driver.get_valid_times(varname, vartype, it)
                    for vt in valid_times:
                        try:
                            self.driver.set_plot_parameters(
                                self, init_time=it, valid_time=vt)
                        except (IOError, ValueError) as ex:
                            logging.debug("WARNING: unsuccessfully examined data for "
                                          "init time %s, valid time %s.. trying next "
                                          "time.", it, vt)
                            logging.debug(u"(Error message: %s %S )", type(ex), ex)
                        else:
                            successful = True
                            break

            # If we're still not successful..
            if not successful:
                logging.error("ERROR: cannot determine whether there's a "
                              "vertical coordinate.. something is wrong with "
                              "this dataset!!")
                return []

            # Get a list of the available levels.
            if self.driver.vert_data is not None:
                return [u"{:}".format(lvl) for lvl in sorted({float(_x) for _x in self.driver.vert_data})]
            else:
                return []
        else:
            logging.debug("Only surface layers or no driver")
            return []

    def get_elevation_units(self):
        """Returns the units of the elevation values.
        """
        if self.driver is not None:
            return self.driver.vert_units
        else:
            return ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

    mslib.mswms.mswms
    ~~~~~~~~~~~~~~~~~

    The module can be run with the Python PASTE framework as a stand-alone
    server (simply execute this file with Python).

    This file is part of mss.

    :copyright: Copyright 2016 Reimar Bauer
    :copyright: Copyright 2016-2018 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import print_function
from __future__ import absolute_import

import paste.httpserver
import argparse
import logging
import sys

from mslib import __version__
from mslib.mswms.wms import mss_wms_settings, mss_wms_auth
from mslib.utils import setup_logging


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", help="show version", action="store_true", default=False)
    parser.add_argument("--host", help="hostname",
                        default="127.0.0.1", dest="host")
    parser.add_argument("--port", help="port", dest="port", default="8081")
    parser.add_argument("--threadpool", help="threadpool", dest="use_threadpool", action="store_true", default=False)
    parser.add_argument("--debug", help="show debugging log messages on console", action="store_true", default=False)
    parser.add_argument("--logfile", help="If set to a name log output goes to that file", dest="logfile",
                        default=None)
    args = parser.parse_args()

    if args.version:
        print("***********************************************************************")
        print("\n            Mission Support System (mss)\n")
        print("***********************************************************************")
        print("Documentation: http://mss.rtfd.io")
        print("Version:", __version__)
        sys.exit()

    setup_logging(args)

    from mslib.mswms.wms import application
    if mss_wms_settings.__dict__.get('enable_basic_http_authentication', False):
        logging.debug("Enabling basic HTTP authentication. Username and "
                      "password required to access the service.")

        from paste.auth.basic import AuthBasicHandler
        import hashlib

        realm = 'Mission Support Web Map Service'

        def authfunc(environ, username, password):
            for u, p in mss_wms_auth.allowed_users:
                if (u == username) and (p == hashlib.md5(password).hexdigest()):
                    return True
            return False

        application = AuthBasicHandler(application, realm, authfunc)
    logging.info(u"Configuration File: '{}'".format(mss_wms_settings.__file__))
    paste.httpserver.serve(application, host=args.host, port=args.port, use_threadpool=args.use_threadpool)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""

    mslib.mswms.dataaccess
    ~~~~~~~~~~~~~~~~~~~~~~

    This module provides functions to access data

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2018 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import division

from abc import ABCMeta, abstractmethod
import itertools
import os
import logging
import netCDF4
import numpy as np

from mslib import netCDF4tools
from future.utils import with_metaclass


class NWPDataAccess(with_metaclass(ABCMeta, object)):
    """Abstract superclass providing a framework to let the user query
       in which data file a given variable at a given time can be found.

    The class provides the method get_filename(). It derives filenames from
    CF variable names, initialisation and valid times.q
    The method get_datapath() provides the root path where the data
    can be found.

    In subclasses, the protected method _determine_filename() must be
    implemented.
    """

    def __init__(self, rootpath):
        """Constructor takes the path of the data directory.
        """
        self._root_path = rootpath
        self._modelname = ""

    @abstractmethod
    def setup(self):
        """Checks for existing files etc. and sets up the class. Called by
           server whenever a client requests a current capability document.
        """
        pass

    def get_filename(self, variable, vartype, init_time, valid_time,
                     fullpath=False):
        """Get the filename of the file in which a given variable at
           a given time can be found.

        Arguments:
        variable -- string with CF name of variable
        vartype -- string specifying the type of the variable (model specific).
                   For example, can be ml (model level), pl (pressure level),
                   or sfc (surface) for ECMWF data.
        init_time -- datetime object with initialisation time of forecast run
        valid_time -- datetime object with valid time of forecast
        fullpath -- if True, the complete path to the file will be returned.
                    Default is False, only the filename will be returned.
        """
        filename = self._determine_filename(variable, vartype,
                                            init_time, valid_time)
        if fullpath:
            return os.path.join(self._root_path, filename)
        else:
            return filename

    @abstractmethod
    def _determine_filename(self, variable, vartype, init_time, valid_time):
        """Must be overwritten in subclass. Determines the filename
           (without path) of the variable <variable> at the forecast
           timestep specified by init_time and valid_time.
        """
        pass

    def get_datapath(self):
        """Return the path to the data directory.
        """
        return self._root_path

    @abstractmethod
    def get_all_datafiles(self):
        """Return a list of all available data files.
        """
        pass

    @abstractmethod
    def get_init_times(self):
        """Return a list of available forecast init times (base times).
        """
        pass

    _mfDatasetArgsDict = {}

    def mfDatasetArgs(self):
        """Returns additional keyword for the MFDatasetCommonDims instance that
           handles the input data of this dataset. See the MFDatasetCommonDims
           documentation for further details.
           Mainly provided as a workaround for numerical inaccuracies introduced
           to the NetCDF files by netcdf-java 4.3.
           (mr, 16Oct2012)
        """
        return self._mfDatasetArgsDict


class DefaultDataAccess(NWPDataAccess):
    """
    Subclass to NWPDataAccess for accessing properly constructed NetCDF files
    Constructor needs information on domain ID.
    """

    # Workaround for the numerical issue concering the lon dimension in
    # NetCDF files produced by netcdf-java 4.3..
    _mfDatasetArgsDict = {"skipDimCheck": ["lon"]}

    def __init__(self, rootpath, domain_id):
        NWPDataAccess.__init__(self, rootpath)
        self._domain_id = domain_id
        self._available_files = None
        self._filetree = None

    def _determine_filename(self, variable, vartype, init_time, valid_time):
        """Determines the name of the ECMWF data file the contains
           the variable <variable> of the forecast specified by
           init_time and valid_time.
        """
        assert self._filetree is not None, "filetree is None. Forgot to call setup()?"
        try:
            return self._filetree[vartype][init_time][variable][valid_time]
        except KeyError as ex:
            self.setup()
            try:
                return self._filetree[vartype][init_time][variable][valid_time]
            except KeyError as ex:
                logging.error("Could not identify filename. %s %s %s %s %s %s",
                              variable, vartype, init_time, valid_time, type(ex), ex)
                raise ValueError(u"variable type {} not available for variable {}"
                                 .format(vartype, variable))

    def setup(self):
        # Get a list of the available data files.
        self._available_files = [
            _filename for _filename in os.listdir(self._root_path) if self._domain_id in _filename]
        logging.info("Files identified for domain '%s': %s",
                     self._domain_id, self._available_files)

        self._filetree = {}
        elevations = {}

        # Build the tree structure.
        for filename in self._available_files:
            logging.info("Opening candidate '%s'", filename)
            try:
                with netCDF4.Dataset(os.path.join(self._root_path, filename)) as dataset:

                    time_name, time_var = netCDF4tools.identify_CF_time(dataset)
                    init_time = netCDF4tools.num2date(0, time_var.units)
                    valid_times = netCDF4tools.num2date(time_var[:], time_var.units)
                    lat_name, lat_var, lon_name, lon_var = netCDF4tools.identify_CF_lonlat(dataset)
                    vert_name, vert_var, _, _, vert_type = netCDF4tools.identify_vertical_axis(dataset)

                    if len(time_var.dimensions) != 1 or time_var.dimensions[0] != time_name:
                        logging.error("Skipping file '%s': problem with time coordinate variable", filename)
                        continue
                    if len(lat_var.dimensions) != 1 or lat_var.dimensions[0] != lat_name:
                        logging.error("Skipping file '%s': problem with latitude coordinate variable", filename)
                        continue
                    if len(lon_var.dimensions) != 1 or lon_var.dimensions[0] != lon_name:
                        logging.error("Skipping file '%s': problem with longitude coordinate variable", filename)
                        continue

                    if vert_type != "sfc":
                        if vert_type in elevations:
                            if len(vert_var[:]) != len(elevations[vert_type]["levels"]):
                                logging.error("Skipping file '%s': number of vertical levels does not fit to levels of "
                                              "previous file '%s'.", filename, elevations[vert_type]["filename"])
                                continue

                            if not np.allclose(vert_var[:], elevations[vert_type]["levels"]):
                                logging.error("Skipping file '%s': vertical levels do not fit to levels of previous "
                                              "file '%s'.", filename, elevations[vert_type]["filename"])
                                continue
                        else:
                            elevations[vert_type] = {"filename": filename, "levels": vert_var[:]}

                    standard_names = []
                    for ncvarname, ncvar in dataset.variables.items():
                        if hasattr(ncvar, "standard_name"):
                            if (len(ncvar.dimensions) >= 3 and (
                                    ncvar.dimensions[0] != time_name or
                                    ncvar.dimensions[-2] != lat_name or
                                    ncvar.dimensions[-1] != lon_name)):
                                logging.error("Skipping variable '%s' in file '%s': Incorrect order of dimensions",
                                              ncvarname, filename)
                                continue
                            if len(ncvar.shape) == 4 and vert_name in ncvar.dimensions:
                                standard_names.append(ncvar.standard_name)
                            elif len(ncvar.shape) == 3 and vert_type == "sfc":
                                standard_names.append(ncvar.standard_name)
            except IOError as ex:
                logging.error("Skipping file '%s' (%s: %s)", filename, type(ex), ex)
                continue

            logging.info("File '%s' identified as '%s' type", filename, vert_type)
            logging.info("Found init time '%s', %s valid_times and %s standard_names",
                         init_time, len(valid_times), len(standard_names))
            if len(valid_times) == 0 or len(standard_names) == 0:
                logging.error(
                    "Something is wrong with this file... valid_times='%s' standard_names='%s'",
                    valid_times, standard_names)
            else:
                logging.debug("valid_times='%s' standard_names='%s'", valid_times, standard_names)

            leaf = self._filetree.setdefault(vert_type, {}).setdefault(init_time, {})
            for standard_name in standard_names:
                var_leaf = leaf.setdefault(standard_name, {})
                for valid_time in valid_times:
                    if valid_time in var_leaf:
                        logging.warning(
                            "some data was found twice! vartype='%s' init_time='%s' standard_name='%s' "
                            "valid_time='%s' first_file='%s' second_file='%s'",
                            vert_type, init_time, standard_name, valid_time, var_leaf[valid_time], filename)
                    else:
                        var_leaf[valid_time] = filename

        return self._filetree

    def get_init_times(self):
        """Returns a list of available forecast init times (base times).
        """
        init_times = set(itertools.chain.from_iterable(
            self._filetree[_x].keys() for _x in self._filetree))
        return sorted(init_times)

    def get_valid_times(self, variable, vartype, init_time):
        """Returns a list of available valid times for the specified
           variable at the specified init time.
        """
        try:
            return sorted(self._filetree[vartype][init_time][variable].keys())
        except KeyError as ex:
            logging.error("Could not find times! %s %s", type(ex), ex)
            return []

    def get_all_valid_times(self, variable, vartype):
        """Similar to get_valid_times(), but returns the combined valid times
           of all available init times.
        """
        all_valid_times = []
        if vartype not in self._filetree:
            return []
        for init_time in self._filetree[vartype]:
            if variable in self._filetree[vartype][init_time]:
                all_valid_times.extend(list(self._filetree[vartype][init_time][variable].keys()))
        return sorted(set(all_valid_times))

    def get_all_datafiles(self):
        """Return a list of all available data files.
        """
        return self._available_files



# -*- coding: utf-8 -*-
"""

    mslib.msui.timeseriesview
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Window for the display of trajectory time series.


    To better understand of the code, look at the 'ships' example from
    chapter 14/16 of 'Rapid GUI Programming with Python and Qt: The
    Definitive Guide to PyQt Programming' (Mark Summerfield).

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2018 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from mslib.msui.mss_qt import QtGui  # Qt bindings

from mslib.msui.mss_qt import ui_timeseriesview_window as ui
from mslib.msui.viewwindows import MSSMplViewWindow
from mslib.msui.icons import icons


class MSSTimeSeriesViewWindow(MSSMplViewWindow, ui.Ui_TimeSeriesViewWindow):
    """PyQt window implementing a time series view.
    """
    name = "Time Series View"

    def __init__(self, parent=None):
        """Set up user interface, connect signal/slots.
        """
        super(MSSTimeSeriesViewWindow, self).__init__(parent)
        self.setupUi(self)
        self.setWindowIcon(QtGui.QIcon(icons('64x64')))

    def setIdentifier(self, identifier):
        super(MSSTimeSeriesViewWindow, self).setIdentifier(identifier)
        self.mpl.canvas.setIdentifier(identifier)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'msui/ui_performance_settings.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_PerformanceSettingsDialog(object):
    def setupUi(self, PerformanceSettingsDialog):
        PerformanceSettingsDialog.setObjectName("PerformanceSettingsDialog")
        PerformanceSettingsDialog.resize(319, 241)
        self.buttonBox = QtWidgets.QDialogButtonBox(PerformanceSettingsDialog)
        self.buttonBox.setGeometry(QtCore.QRect(30, 200, 271, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.dteTakeoffTime = QtWidgets.QDateTimeEdit(PerformanceSettingsDialog)
        self.dteTakeoffTime.setGeometry(QtCore.QRect(160, 100, 141, 22))
        self.dteTakeoffTime.setObjectName("dteTakeoffTime")
        self.label = QtWidgets.QLabel(PerformanceSettingsDialog)
        self.label.setGeometry(QtCore.QRect(20, 100, 91, 16))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.dsbTakeoffWeight = QtWidgets.QDoubleSpinBox(PerformanceSettingsDialog)
        self.dsbTakeoffWeight.setGeometry(QtCore.QRect(160, 130, 141, 22))
        self.dsbTakeoffWeight.setDecimals(0)
        self.dsbTakeoffWeight.setMaximum(9999999.0)
        self.dsbTakeoffWeight.setSingleStep(1000.0)
        self.dsbTakeoffWeight.setObjectName("dsbTakeoffWeight")
        self.dsbFuel = QtWidgets.QDoubleSpinBox(PerformanceSettingsDialog)
        self.dsbFuel.setGeometry(QtCore.QRect(160, 160, 141, 22))
        self.dsbFuel.setDecimals(0)
        self.dsbFuel.setMaximum(9999999.0)
        self.dsbFuel.setSingleStep(1000.0)
        self.dsbFuel.setObjectName("dsbFuel")
        self.label_2 = QtWidgets.QLabel(PerformanceSettingsDialog)
        self.label_2.setGeometry(QtCore.QRect(20, 130, 121, 16))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.label_3 = QtWidgets.QLabel(PerformanceSettingsDialog)
        self.label_3.setGeometry(QtCore.QRect(20, 160, 101, 16))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.cbShowPerformance = QtWidgets.QCheckBox(PerformanceSettingsDialog)
        self.cbShowPerformance.setGeometry(QtCore.QRect(20, 20, 141, 20))
        self.cbShowPerformance.setObjectName("cbShowPerformance")
        self.pbLoadPerformance = QtWidgets.QPushButton(PerformanceSettingsDialog)
        self.pbLoadPerformance.setGeometry(QtCore.QRect(210, 50, 93, 28))
        self.pbLoadPerformance.setObjectName("pbLoadPerformance")
        self.lbAircraftName = QtWidgets.QLabel(PerformanceSettingsDialog)
        self.lbAircraftName.setGeometry(QtCore.QRect(80, 60, 121, 16))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.lbAircraftName.setFont(font)
        self.lbAircraftName.setObjectName("lbAircraftName")
        self.label_5 = QtWidgets.QLabel(PerformanceSettingsDialog)
        self.label_5.setGeometry(QtCore.QRect(20, 60, 51, 16))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label_5.setFont(font)
        self.label_5.setObjectName("label_5")

        self.retranslateUi(PerformanceSettingsDialog)
        self.buttonBox.accepted.connect(PerformanceSettingsDialog.accept)
        self.buttonBox.rejected.connect(PerformanceSettingsDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(PerformanceSettingsDialog)

    def retranslateUi(self, PerformanceSettingsDialog):
        _translate = QtCore.QCoreApplication.translate
        PerformanceSettingsDialog.setWindowTitle(_translate("PerformanceSettingsDialog", "PerformanceSettingsDialog"))
        self.label.setText(_translate("PerformanceSettingsDialog", "Take off time"))
        self.label_2.setText(_translate("PerformanceSettingsDialog", "Takeoff weight (lb)"))
        self.label_3.setText(_translate("PerformanceSettingsDialog", "Fuel (lb)"))
        self.cbShowPerformance.setText(_translate("PerformanceSettingsDialog", "Show Performance"))
        self.pbLoadPerformance.setText(_translate("PerformanceSettingsDialog", "Load"))
        self.lbAircraftName.setText(_translate("PerformanceSettingsDialog", "Dummy"))
        self.label_5.setText(_translate("PerformanceSettingsDialog", "Aircraft:"))




# -*- coding: utf-8 -*-
"""

    mslib.msui.trajectory_item_tree
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tree model for trajectory items loaded into the trajectory tool and drawn
    on any views.

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2018 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import division

from builtins import str

import datetime
import logging
import os
from mslib.utils import config_loader
from mslib.msui import MissionSupportSystemDefaultConfig as mss_default
# related third party imports
from mslib.msui.mss_qt import QtCore, QtGui
import numpy

try:
    import nappy
    HAVE_NAPPY = True
except ImportError:
    logging.warning("*** NAppy is not available. You will not be able to read NASA Ames files. ***")
    HAVE_NAPPY = False

from mslib.msui import lagranto_output_reader


class LagrantoTreeModelUnsupportedOperationError(Exception):
    """Exception class to handle wrong method arguments.
    """
    pass


class AbstractLagrantoDataItem(object):
    """Base class for all trajectory instances that are loaded into a
       trajectory tool.

    An AbstractLagrantoDataItem is a node in the tree of trajectory items
    that are drawn on any view. Hence, it has children and a parent. The
    class provides all methods required by Qt for display in a QTreeView.
    It also provides methods common to all trajectory items, including
    in which views it is visible.
    """

    def __init__(self, name, visible, parent=None):
        # Parent and children of the node in the tree.
        self.parentItem = parent
        self.childItems = []
        self.itemName = name
        self.hasMetadata = False

        # gxElements stores properties of the item, including colour,
        # line style and thickness, and plot instances of any view
        # (each view is free to store whatever it likes).
        self.gxElements = {}
        # List of views in which the item is visible. Each view is stored
        # by its identification string.
        self.views = []

        # If this item has a parent, register it as a child in the parent
        # node.
        if parent is not None:
            parent.appendChild(self)

    def appendChild(self, item):
        self.childItems.append(item)

    def child(self, row):
        return self.childItems[row]

    def childCount(self):
        return len(self.childItems)

    def treeViewData(self, column):
        """Return string data for display in the tree view.
        """
        if column == 0:
            # Item name.
            return self.itemName
        elif column == 1:
            # Connected views.
            return '; '.join(self.views)
        elif column == 2:
            # Item line properties (if not applicable return empty string).
            try:
                return str(self.gxElements['general']['colour']) + '/' + str(
                    self.gxElements['general']['linestyle']) + '/' + str(
                        self.gxElements['general']['linewidth'])
            except Exception as ex:
                logging.debug(u"caught a wildcard Exception: %s, %s", type(ex), ex)
                return ''
        elif column == 3:
            # Item markers.
            s = ''
            try:
                s += u'time({})'.format(self.gxElements['general']['timeMarkerInterval'].strftime('%H:%M'))
            except Exception as ex:
                logging.debug(u"caught a wildcard Exception: %s, %s", type(ex), ex)
            return s
        else:
            return ''

    def parent(self):
        return self.parentItem

    def row(self):
        if self.parentItem is not None:
            return self.parentItem.childItems.index(self)
        else:
            return 0

    def setVisibleInView(self, view_id, visible):
        """Set the visibility of this item in the view with the identifier
           <view_id>.
        """
        if visible:
            if view_id not in self.views:
                self.views.append(view_id)
                if self.parentItem:
                    self.parentItem.setVisibleInView(view_id, visible)
        else:
            if view_id in self.views:
                self.views.remove(view_id)

    def isVisible(self, view_id):
        """Is this item visible in the view with identifier <view_id>?
        """
        #         logging.debug("%s is visible on %s: %s" % (self.itemName,
        #                                                    view_id,
        #                                                    view_id in self.views))
        return view_id in self.views

    def getName(self):
        return self.itemName

    def isRoot(self):
        """Return True if item is the root item of a tree (ie., it has no
           parent).
        """
        return not self.parentItem

    def setGxElementProperty(self, element, eproperty, value):
        """
        """
        if element not in self.gxElements:
            self.gxElements[element] = {}
        self.gxElements[element][eproperty] = value

    def getGxElements(self):
        return self.gxElements

    def getGxElementProperty(self, element, eproperty):
        return self.gxElements[element][eproperty]

    def hasMetadata(self):
        return self.hasMetadata

    def getMetadata(self):
        """Return metadata dictionary of the item.

        This method is ABSTRACT and has to be implemented in the derived
        classes.
        """
        raise NotImplementedError("Abstract AbstractLagrantoDataItem.getMetadata called.")

    def getMetadataValue(self, key):
        """Return the value belonging to a metadata key.

        This method is ABSTRACT and has to be implemented in the derived
        classes.
        """
        raise NotImplementedError("Abstract AbstractLagrantoDataItem.getMetadataValue called.")


class LagrantoMapItem(AbstractLagrantoDataItem):
    """
    """

    def __init__(self, name, visible, parent=None):
        """
        """
        super(LagrantoMapItem, self).__init__(name, visible, parent)
        self.timeVariableChild = None
        self.lonVariableChild = None
        self.latVariableChild = None
        self.pressureVariableChild = None
        self.timeMarkerIndexes = None

        self.setGxElementProperty("general", "colour", None)
        self.setGxElementProperty("general", "linestyle", None)
        self.setGxElementProperty("general", "linewidth", None)
        self.setGxElementProperty("general", "timeMarkerInterval", None)

    def treeViewData(self, column):
        """Return string data for display in the tree view.

        Overrides AbstractLagrantoDataItem.treeViewData() to provide data
        for column 4 (start coordinates).
        """
        if column < 4:
            return AbstractLagrantoDataItem.treeViewData(self, column)
        elif column == 4:
            # Start coordinates from first elements of lon/lat/p variables.
            try:
                return "{:.2f}, {:.2f}, {:.1f}".format(
                    self.lonVariableChild.getVariableData()[0],
                    self.latVariableChild.getVariableData()[0],
                    self.pressureVariableChild.getVariableData()[0])
            except Exception as ex:
                logging.debug("Wildecard Exception %s - %s.", type(ex), ex)
                return ''
        else:
            return ''

    def getTimeVariable(self):
        """Return the child item that holds the time variable.
        """
        return self.timeVariableChild

    def getLonVariable(self):
        """Return the child item that holds the longitude variable.
        """
        return self.lonVariableChild

    def getLatVariable(self):
        """Return the child item that holds the latitude variable.
        """
        return self.latVariableChild

    def getPressureVariable(self):
        """Return the child item that holds the pressure variable.
        """
        return self.pressureVariableChild

    def getTimeMarkerIndexes(self):
        """Return an array containing the indexes of the time marker points.
        """
        return self.timeMarkerIndexes

    def setGxElementProperty(self, element, eproperty, value):
        """Overloads AbstractLagrantoDataItem.setGxElementProperty(). Calls
           the inherited method, and if the 'general' property
           'timeMarkerInterval' is set the data array indexes of the
           time markers are computed with __computeTimeMarkerPoints().
        """
        AbstractLagrantoDataItem.setGxElementProperty(self, element,
                                                      eproperty, value)
        if element == "general" and eproperty == "timeMarkerInterval":
            self.__computeTimeMarkerIndexes(value)
            try:
                self.timeSeriesPlotter.update()
            except Exception as ex:
                logging.debug("Wildecard Exception %s - %s.", type(ex), ex)
                pass

    def __computeTimeMarkerIndexes(self, requestedInterval):
        """Computes the indexes for the variable arrays at which time markers
           should be placed.

        The argument 'requestedInterval' is a datetime object representing the
        time interval between the marker points (e.g. three hours). If the
        requested time interval is smaller than the time distance between two
        points in the time variable, the interval value that's being used is
        set to this minumum value (this can be checked by comparing the
        requested value to the value that's contained in gxElements/general/
        timeMarkerInterval after this method has been called.

        A numpy.array containing the indexes (that can be used to index the
        lon/lat variables, for instance) is accessible via
        getTimeMarkerIndexes() after this method has been called.
        """
        #
        # Convert the requested interval from the datetime object to an int
        # value representing seconds.
        try:
            requestedInterval = requestedInterval.hour * 3600 + requestedInterval.minute * 60
        except Exception as ex:
            logging.debug("Wildecard Exception %s - %s.", type(ex), ex)
            pass

        #
        # If the requested time interval is zero or a None-tpye is passed,
        # no marker points will be computed.
        if not requestedInterval or requestedInterval == 0:
            self.timeMarkerIndexes = None
            self.gxElements["general"]["timeMarkerInterval"] = None
            return
        #
        # If this object (self) has no time variable skip the computation
        # of the indexes. The value of 'requestedInterval' will still be
        # stored in gxElements, so that it can be handed down to the
        # children.
        if not self.timeVariableChild:
            self.timeMarkerIndexes = None
            return

        #
        # Get the time variable of this object (which is returned in hours)
        # and convert to seconds (to get better accuracy in searchsorted()
        # below?!? Maybe it's not needed..).
        timeDataInSeconds = 3600. * self.timeVariableChild.getVariableData()
        #
        # Determine the mininum possible time interval (difference between to
        # data points in the time variable).
        minInterval = timeDataInSeconds[1] - timeDataInSeconds[0]

        #
        # The interval we'll use is the larger value of the two.
        interval = max(requestedInterval, minInterval)
        #
        # If we're not using the requested interval store the new interval value
        # to gxElements.
        if interval != requestedInterval:
            self.gxElements["general"]["timeMarkerInterval"].replace(hour=0,
                                                                     minute=0)
            self.gxElements["general"]["timeMarkerInterval"] += datetime.timedelta(seconds=interval)

        #
        # Compute the time marker times: arange from the first to the last data
        # point in the time variable, spacing is 'interval'.
        # Use min/max functions to handle backwards trajectories. In this case
        # the last element of timeDateInSeconds is smaller than the first,
        # which would lead to an empty list from arange.
        if timeDataInSeconds[-1] < 0:
            timeDataInSeconds = timeDataInSeconds[::-1]
        markerTimesInSeconds = numpy.arange(timeDataInSeconds[0],
                                            timeDataInSeconds[-1],
                                            interval)
        #
        # Find the indexes in the time variable array that correspond to these
        # times.
        self.timeMarkerIndexes = timeDataInSeconds.searchsorted(
            markerTimesInSeconds)

    def translateQueryString(self, qstring):
        """Translate a user query string containing %lon, %lat, %pres and %meta
           identifiers into a Python statement that can be evaluated with
           eval().

        The function replaces the %-identifiers by the appropriate Python
        statements that access the data.

        Arguments:
        qstring -- user query string.
        """
        qstring = qstring.replace('%lon',
                                  'self.lonVariableChild.getVariableData()[0]')
        qstring = qstring.replace('%lat',
                                  'self.latVariableChild.getVariableData()[0]')
        qstring = qstring.replace('%pres',
                                  'self.pressureVariableChild.getVariableData()[0]')
        qstring = qstring.replace('%meta', 'self.getMetadataValue')
        return qstring

    def query(self, qstring):
        """Evaluate the translated query string qstring. Note that the query
           must be translated into a Python statement by using
           'translateQueryString()' before calling this function.

        Arguments:
        qstring -- user query string.
        """
        return eval(qstring)


class FlightTrackItem(LagrantoMapItem):
    """Holds flight track data stored in a NASA Ames file.
    """

    def __init__(self, nasFileName, visible, parent=None):
        """Constructor.

        Arguments:
        nasFileName -- full path NASA Ames file
        visible,
            parent  -- inherited from LagrantoMapItem
        """
        super(FlightTrackItem, self).__init__(
            os.path.basename(nasFileName), visible, parent)
        self.setGxElementProperty("general", "colour", "blue")
        self.setGxElementProperty("general", "linestyle", "-")
        self.setGxElementProperty("general", "linewidth", 1)

        self.nasFileName = nasFileName
        self.nafile = None
        self.__readNasaAmesFile()

        if self.nafile is None:
            logging.error("could not read NASA Ames file")
            raise RuntimeError("could not read NASA Ames file")

        #
        # Add all variable names contained in self.data as child nodes, so
        # that they are added to the tree view.
        variableNames = self.nafile.VNAME
        for variable in variableNames:
            FlightTrackVariableItem(variable, False, self)

        self.timeVariableChild = None
        for identifier in [self.nafile.XNAME[0]]:
            for item in self.childItems:
                if item.getName().find(identifier) >= 0:
                    self.timeVariableChild = item
                    logging.debug(u"identified time variable <%s>", item.getName())
                    #
                    # The time variable has to be modified a bit: convert
                    # seconds to hours and change the name correspondingly.
                    index = self.nafile.VNAME.index(item.getName())
                    self.nafile.V[index, :] = (self.nafile.V[index, :] / 3600.)
                    self.nafile.VNAME[index] = "(1/3600) * " + self.nafile.VNAME[index]
                    item.itemName = self.nafile.VNAME[index]
                    break
            if self.timeVariableChild is not None:
                break

        self.lonVariableChild = None
        for identifier in config_loader(dataset="traj_nas_lon_identifier", default=mss_default.traj_nas_lon_identifier):
            for item in self.childItems:
                if item.getName().upper().find(identifier) >= 0:
                    self.lonVariableChild = item
                    logging.debug(u"identified longitude variable <%s> with "
                                  u"identifier <%s>", item.getName(), identifier)
                    break
            if self.lonVariableChild is not None:
                break

        self.latVariableChild = None
        for identifier in config_loader(dataset="nas_lat_identifier", default=mss_default.traj_nas_lat_identifier):
            for item in self.childItems:
                if item.getName().upper().find(identifier) >= 0:
                    self.latVariableChild = item
                    logging.debug(u"identified latitude variable <%s> with "
                                  u"identifier <%s>", item.getName(), identifier)
                    break
            if self.latVariableChild is not None:
                break

        self.pressureVariableChild = None
        for identifier in config_loader(dataset="traj_nas_p_identifier", default=mss_default.traj_nas_p_identifier):
            for item in self.childItems:
                if item.getName().upper().find(identifier) >= 0:
                    self.pressureVariableChild = item
                    logging.debug(u"identified pressure variable <%s> with "
                                  u"identifier <%s>", item.getName(), identifier)
                    break
            if self.pressureVariableChild is not None:
                break

    def __readNasaAmesFile(self):
        """Read the NASA Ames file 'self.nasFileName'.
        """
        if not HAVE_NAPPY:
            self.nafile = None
            return
        #
        # Open the NASA Ames file and read the data.
        try:
            self.nafile = nappy.openNAFile(self.nasFileName)
            self.nafile.readData()
        except TypeError as ex:  # catch TypeError as nappy itself triggers Exception when raising
            self.nafile = None
            logging.error(u"%s %s", type(ex), ex)
            return
        #
        # Convert variable array of nafile from 'list' to 'NumPy array' for
        # faster access (the array can be large!).
        #  -- Should be moved to nappy, not a great solution..
        self.nafile.V = numpy.array(self.nafile.V)
        # independent variable
        self.nafile.X = numpy.array(self.nafile.X)

        # Convert to masked array to correctly handle missing values.
        Vtemp = []
        for i in range(self.nafile.V.shape[0]):
            Vtemp.append(numpy.ma.masked_values(self.nafile.V[i],
                                                self.nafile.VMISS[i]))
        # Append independent variable to list of variables. Independent
        # variables don't have a missing value, hence the dummy -9999.. here.
        # TODO: This should be revised; the individual FlightTrackVariableItem shouldn't
        #       access their parent's 'self.nafile.V' attribute to access the data;
        #       instead they should own their own data array.
        Vtemp.append(numpy.ma.masked_values(self.nafile.X,
                                            -99999999))
        self.nafile.VNAME.append(self.nafile.XNAME[0])

        self.nafile.V = numpy.ma.masked_array(Vtemp)

    def getMetadataValue(self, key):
        """Return the value belonging to a metadata key.

        Implements AbstractLagrantoDataItem.getMetadataValue(). Always returns
        None as flight track items have no metadata.
        """
        return None

    def getStartTime(self):
        """Return the start time of the trajectory as a datetime object.
        """
        # Get the date from the NASA Ames file as a list [YYYY, MM, DD].
        nas_date = self.nafile.getFileDates()[0]
        return datetime.datetime(nas_date[0], nas_date[1], nas_date[2])


class LagrantoOutputItem(LagrantoMapItem):
    """Holds all data stored in a Lagranto output directory.
    """

    def __init__(self, path, visible, parent=None):
        """Constructor.

        Arguments:
        path -- full path to directory in which Lagranto output files are
                stored
        visible,
        parent -- inherited from LagrantoMapItem
        """
        super(LagrantoOutputItem, self).__init__(
            os.path.basename(os.path.normpath(path)), visible, parent)

        self.path = path
        self.loutput = None
        self.__readLagrantoOutput()

    def __readLagrantoOutput(self):
        """
        """
        #
        # Read the data via a LagrantoOutputReader object.
        self.loutput = lagranto_output_reader.LagrantoOutputReader(self.path)
        #
        # Instantiate a new TrajectoryItem for each trajectory read by
        # LagrantoOutputReader.
        for i, (trajectory, metadata) in enumerate(zip(self.loutput.data,
                                                       self.loutput.meta)):
            trname = "{:04d} ".format(i)
            if "startcoordinates" in metadata:
                trname += str(
                    [u"{:.2f}".format(r) for r in metadata["startcoordinates"]]).replace('\'', '')
            TrajectoryItem(trname, True, self, trajectory, metadata)


class TrajectoryItem(LagrantoMapItem):
    """Holds the data from an individual trajectory computed by Lagranto.

    Data is accessible as TrajectoryItem.data['varname']; all available
    variables can be queried with TrajectoryItem.data.keys().
    """

    def __init__(self, trajectoryname, visible, parent=None, trdata=None,
                 trmetadata=None):
        """Constructor.

        Arguments:
        trajectoryname -- name of the trajectory
        visible,
            parent -- inherited from LagrantoMapItem
        trdata   -- dictionary with trajectory data (trdata['varname'] ==
                    NumPy array)
        trmetadata -- trajectory metadata dictionary, as returned by
                      LagrantoOutputReader
        """
        super(TrajectoryItem, self).__init__(trajectoryname, visible, parent)
        self.setGxElementProperty("general", "colour", "red")
        self.setGxElementProperty("general", "linestyle", "-")
        self.setGxElementProperty("general", "linewidth", 1)
        self.data = trdata
        self.metadata = trmetadata
        self.hasMetadata = True

        #
        # Add all variable names contained in self.data as child nodes, so
        # that they are added to the tree view.
        for variable in self.data:
            #
            # Per default, plot only variables listed here in a new time
            # series window.
            visible = variable in ['p', 'T', 'TH', 'RH']
            newItem = TrajectoryVariableItem(variable, visible, self)
            #
            # Remember the items that store time, lon, lat, pressure variables
            # for later access with getTimeVariable() etc.
            if variable == "time":
                self.timeVariableChild = newItem
                self.timeVariableChild.itemName += " [hr since " + \
                                                   self.getStartTime().strftime("%Y-%m-%d %H:%M UTC") + "]"
                self.data[self.timeVariableChild.itemName] = self.data[variable]
            elif variable == "lon":
                self.lonVariableChild = newItem
            elif variable == "lat":
                self.latVariableChild = newItem
            elif variable == "p":
                self.pressureVariableChild = newItem

    def treeViewData(self, column):
        """Return string data for display in the tree view.

        Overrides LagrantoMapItem.treeViewData() to provide data
        for columns 5 and 6 (start time & duration, metadata).
        """
        if column < 5:
            return LagrantoMapItem.treeViewData(self, column)
        elif column == 5:
            return "{} for {:f} hrs" \
                .format(self.getStartTime().strftime("%Y-%m-%d %H:%M UTC"),
                        self.timeVariableChild.getVariableData()[-1])
        elif column == 6:
            s = ''
            for key, value in self.metadata.items():
                if key not in ["starttime", "file", "starttime_filename",
                               "startcoordinates", "duration"]:
                    s += u"{} = {}, ".format(key, str(value))
            return s

    def getStartTime(self):
        """Return the start time of the trajectory as a datetime object.
        """
        try:
            return self.metadata["starttime"]
        except KeyError:
            return self.metadata["starttime_filename"]

    def getMetadata(self):
        """Return the metadata dictionary.

        This method implements AbstractLagrantoDataItem.getMetadata().
        """
        return self.metadata

    def getMetadataValue(self, key):
        """Return the value belonging to a metadata key.

        Implements AbstractLagrantoDataItem.getMetadataValue(). Returns None
        if the key is not contained in the metadata dictionary.
        """
        try:
            return self.metadata[key]
        except KeyError:
            return None


class AbstractVariableItem(AbstractLagrantoDataItem):
    """Tree node for variables contained in a flight track or trajectory file.
       Such variable can be visible on a time series plot, or they can be used
       to colour the track plotted on the map. Variable data are stored in
       flight track and trajectory nodes, and since these data fields can be
       large they shouldn't be copied into the variable nodes. Hence, only
       the name of the variables is stored locally (accessible via
       getVariableName()), and the data have to be made accessible by
       overloading getVariableData() in derived classes. The overloaded methods
       should then access the data structure of their parent and return the
       data, so that 'users' such as LagrantoMap or LagrantoTimeSeriesPlotter
       can use the common interface and not bother about the data is stored
       in respective parent nodes.
    """

    def __init__(self, name, visible, parent=None):
        """The constructor takes the same arguments as AbstractLagrantoDataItem.
           Only the constructor of the superclass is called.
        """
        super(AbstractVariableItem, self).__init__(name, visible, parent)

    def getVariableName(self):
        """Return the name of the variable.
        """
        return self.itemName

    def getVariableData(self):
        """Return the data associated with this variable. This method depends
           on the type of parent (ie. flight track or trajectory) and has
           to be overloaded in the derived class so that efficient data
           access is possible.

        This method is ABSTRACT and has to be implemented in the derived
        classes.
        """
        raise NotImplementedError("Abstract AbstractVariableNameItem.getVariableData called.")

    def exceedsThreshold(self, threshold=0.):
        """Determines time intervals in which the value of the variable
           exceeds a given threshold.

        Returns a list of two-element-lists of form
                        [ [ts1, te1], [ts2, te2], ..],
        where each two-element-list describes a time interval in which the
        value of the variables exceeds the given threshold.

        Arguments:
        threshold -- threshold to be exceeded. Default is 0.
        """
        myData = self.getVariableData()
        timeData = self.parentItem.getTimeVariable().getVariableData()

        aboveThreshold = False
        timeWindows = []
        for i, data in enumerate(myData):
            if not aboveThreshold and data > threshold:
                aboveThreshold = True
                windowStartTime = timeData[i]
            elif aboveThreshold and data <= threshold:
                aboveThreshold = False
                timeWindows.append([windowStartTime, timeData[i]])

        return timeWindows

    def setGxElementProperty(self, element, eproperty, value):
        """Overloads AbstractLagrantoDataItem.setGxElementProperty(). Calls
           the inherited method, and if the 'general' property
           'timeMarkerInterval' is set the data array indexes of the
           time markers are computed with __computeTimeMarkerPoints().
        """
        if element == "general":
            raise LagrantoTreeModelUnsupportedOperationError("Setting " + element + "/" + eproperty +
                                                             " is not supported by AbstractVariableItem.")
        else:
            AbstractLagrantoDataItem.setGxElementProperty(self, element,
                                                          eproperty, value)


class FlightTrackVariableItem(AbstractVariableItem):
    """
    """

    def getVariableData(self):
        """Return the data associated with this variable.

        Implements AbstractVariableItem.getVariableData().
        """
        naVNames = self.parentItem.nafile.VNAME
        index = naVNames.index(self.itemName)
        return self.parentItem.nafile.V[index, :]

    def getGxElementProperty(self, element, eproperty):
        """Return graphics properties (colour etc.) that are stored in the
           'general' element from the parent item.
        """
        if element == "general":
            return self.parentItem.getGxElementProperty(element, eproperty)
        else:
            return AbstractVariableItem.getGxElementProperty(self, element,
                                                             eproperty)


class TrajectoryVariableItem(AbstractVariableItem):
    """
    """

    def getVariableData(self):
        """Return the data associated with this variable.

        Implements AbstractVariableItem.getVariableData().
        """
        return self.parentItem.data[self.itemName]


class LagrantoMapItemsTreeModel(QtCore.QAbstractItemModel):
    """

    .. the model should know nothing about the kind of viewports which
       observe it..

    """

    def __init__(self, data=None, parent=None):
        super(LagrantoMapItemsTreeModel, self).__init__(parent)

        self.header = ['Items', 'Connected Views', 'Line Properties', 'Marker',
                       'Start [lon, lat, p]', 'Start Time', 'Metadata']

        self.rootItem = LagrantoMapItem("All items", True)
        self.rootFlightTracks = LagrantoMapItem("flight tracks", True,
                                                self.rootItem)
        self.rootTrajectories = LagrantoMapItem("trajectories", True,
                                                self.rootItem)

        self.lastChange = "INIT"

    def getRootItem(self):
        return self.rootItem

    def columnCount(self, parent):
        """Return number of columns to be displayed in the tree view.

        Implementation of QAbstractItemModel.columnCount(self,
        QModelIndex parent = QModelIndex()). Here, the number of columns
        is independent of the parent.
        """
        return len(self.header)

    def data(self, index, role):
        """Return data to be displayed in the tree view.
        """
        if not index.isValid():
            return QtCore.QVariant()

        if role != QtCore.Qt.DisplayRole:
            return QtCore.QVariant()

        item = index.internalPointer()

        return QtCore.QVariant(item.treeViewData(index.column()))

    def flags(self, index):
        if not index.isValid():
            return QtCore.Qt.ItemIsEnabled

        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

    def headerData(self, section, orientation, role):
        """Return descriptions of the columns ('sections') in the tree view
           for display in the header.
        """
        #
        # Return value has to be casted as QVariant.
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return QtCore.QVariant(self.header[section])

        return QtCore.QVariant()

    def index(self, row, column, parent):
        """
        Arguments:
        int row, column
        QModelIndex parent
        """
        if not ((0 <= row < self.rowCount(parent)) and (0 <= column < self.columnCount(parent))):
            return QtCore.QModelIndex()

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QtCore.QModelIndex()

    def parent(self, index):
        if not index.isValid():
            return QtCore.QModelIndex()

        childItem = index.internalPointer()
        parentItem = childItem.parent()

        if parentItem == self.rootItem:
            return QtCore.QModelIndex()

        return self.createIndex(parentItem.row(), 0, parentItem)

    def rowCount(self, parent):
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        return parentItem.childCount()

    def addFlightTrack(self, flighttrack):
        """

        Arguments:
        flighttrack -- full path to NASA Ames file that contains the flight
                       track data.
        """
        #
        # Create a new FlightTrackItem as a child of self.rootFlightTracks,
        # create an index to return.
        newFlightTrackItem = FlightTrackItem(flighttrack, True,
                                             self.rootFlightTracks)
        index = self.createIndex(newFlightTrackItem.row(), 0,
                                 newFlightTrackItem)

        #
        # Emit dataChanged() (for LagrantoMap) and layoutChanged() (for
        # QTreeView) signals to inform event listeners about the changes
        # in the data structure. self.lastChange is set to "FLIGHT_TRACK_ADDED"
        # so LagrantoMap can query the kind of change that occured.
        self.lastChange = "FLIGHT_TRACK_ADDED"
        self.dataChanged.emit(index, index)
        self.layoutChanged.emit()

        return index

    def addTrajectoryDirectory(self, path):
        """

        Arguments:
        path -- full path to Lagranto output directory that contains the
                trajectory files (ls* and lsl*).
        """
        #
        # Create a new FlightTrackItem as a child of self.rootFlightTracks,
        # create an index to return.
        newLagrantoOutputItem = LagrantoOutputItem(path, True,
                                                   self.rootTrajectories)
        index = self.createIndex(newLagrantoOutputItem.row(), 0,
                                 newLagrantoOutputItem)

        #
        # Emit dataChanged() (for LagrantoMap) and layoutChanged() (for
        # QTreeView) signals to inform event listeners about the changes
        # in the data structure. self.lastChange is set to "FLIGHT_TRACK_ADDED"
        # so LagrantoMap can query the kind of change that occured.
        self.lastChange = "TRAJECTORY_DIR_ADDED"
        self.dataChanged.emit(index, index)
        self.layoutChanged.emit()

        return index

    def setItemVisibleInView(self, index, view_window, visible,
                             emit_change=True):
        """
        """
        if index.isValid():
            index.internalPointer().setVisibleInView(view_window.identifier,
                                                     visible)
            if emit_change:
                self.lastChange = "VISIBILITY_CHANGE"
                self.dataChanged.emit(index, index)
                self.layoutChanged.emit()

    def setItemVisibleInView_list(self, indices, view_window, visible,
                                  emit_change=True):
        """
        """
        for index in indices:
            if index.isValid():
                index.internalPointer().setVisibleInView(view_window.identifier,
                                                         visible)
            else:
                indices.remove(index)
        if emit_change:
            self.lastChange = "VISIBILITY_CHANGE"
            self.dataChanged.emit(indices[0], indices[-1])
            self.layoutChanged.emit()

    def emitChange(self, index1, index2, mode="GXPROPERTY_CHANGE"):
        self.lastChange = mode
        self.dataChanged.emit(index1, index2)
        self.layoutChanged.emit()

    def getLastChange(self):
        """Returns a description of the last change that has been made
           to the data structure. This method can be called from a listener
           that reacts to the 'dataChanged()' signal if information
           is needed about the type of change that occurred.

        Possible return values (strings) are:
            GXPROPERTY_CHANGE
            VISIBILITY_CHANGE
            TRAJECTORY_DIR_ADDED
            FLIGHT_TRACK_ADDED
            MARKER_CHANGE
        """
        return self.lastChange

    def changeItemGxProperty(self, index, element, eproperty, value):
        """Change a graphics property of the item given by index. Specify
           graphics element, property to be changed and the new value.
           The new value is set and 'dataChanged()' and 'layoutChanged()'
           signals are emitted in order to update a view.
        """
        if index.isValid():
            item = index.internalPointer()
            item.setGxElementProperty(element, eproperty, value)
            self.lastChange = "GXPROPERTY_CHANGE"
            self.dataChanged.emit(index, index)
            self.layoutChanged.emit()

    def changeItemGxProperty_list(self, indices, element, eproperty, value):
        """Change a graphics property of the items given by the list indices.
           Specify graphics element, property to be changed and the new value.
           The new value is set and 'dataChanged()' and 'layoutChanged()'
           signals are emitted in order to update a view.
        """
        for index in indices:
            if index.isValid():
                item = index.internalPointer()
                item.setGxElementProperty(element, eproperty, value)
            else:
                indices.remove(index)
        if len(indices) > 0:
            self.lastChange = "GXPROPERTY_CHANGE"
            self.dataChanged.emit(indices[0], indices[-1])
            self.layoutChanged.emit()

    def setTimeMarker(self, index, interval, emit_change=True):
        """Set the time marker interval property of the item given by the
           index. If the given time interval is smaller than the interval
           between the individual data points of the item, the item will
           automatically set the time interval to this minimum value.
           This method returns the value that has actually been set.
        """
        if index.isValid():
            item = index.internalPointer()
            item.setGxElementProperty("general", "timeMarkerInterval",
                                      interval)
            if emit_change:
                self.lastChange = "MARKER_CHANGE"
                self.dataChanged.emit(index, index)
                self.layoutChanged.emit()

            return item.getGxElementProperty("general", "timeMarkerInterval")

    def selectionFromQuery(self, query, index=None):
        """Translate a selection query string into a Python statement and
           check for each item in the tree if the query is fulfilled.
           Return a QItemSelection of all items that match the query.

        Arguments:
        query -- string of format '%lat >= 20. and %pres >= 500.'. See
                 LagrantoMapItem.translateQueryString() for exact syntax.

        Keyword arguments:
        index -- if the query should only be applied to children of a specific
                 item pass this item's index in this argument. If index is None
                 (default), the query will be applied to all items in the tree.
        """
        if index is not None and index.isValid():
            itemStack = [index.internalPointer()]
        else:
            itemStack = [self.rootItem]

        # Translate the query string into a Python statement that can
        # be evaluated by eval().
        queryTranslation = self.rootItem.translateQueryString(query)

        # Create an instance of a QItemSelection that can store a list of
        # all items that match the query.
        #
        # FROM THE PYQT DOCUMENTATION:
        # A QItemSelection is basically a list of selection ranges, see
        # QItemSelectionRange. QItemSelection saves memory, and avoids
        # unnecessary work, by working with selection ranges rather than
        # recording the model item index for each item in the selection.
        # Generally, an instance of this class will contain a list of
        # non-overlapping selection ranges.
        itemSelection = QtGui.QItemSelection()

        # Traverse the tree:
        # Operate on the items in the stack until the stack is empty.
        while len(itemStack) > 0:
            #
            # Get a new item from the stack.
            item = itemStack.pop()
            #
            # Push all children of the current item onto the stack that are
            # instances of LagrantoMapItem (VariableItems are not plotted on
            # the map and cannot be selected).
            itemStack.extend([child for child in item.childItems
                              if isinstance(child, LagrantoMapItem)])
            #
            # Only flight tracks and trajectories can be queried and selected.
            # Check if the item matches the query.
            if isinstance(item, (FlightTrackItem, TrajectoryItem)) and item.query(queryTranslation):
                # The properties of the current item fulfil the query and
                # the item should be selected. Hence create an index.
                index = self.createIndex(item.row(), 0, item)
                #
                # QtGui.QItemSelectionRange.__init__ (self,
                #     QModelIndex index)
                itemSelectionRange = QtGui.QItemSelectionRange(index)
                itemSelection.append(itemSelectionRange)

        return itemSelection



# -*- coding: utf-8 -*-
"""

    mslib.msui.kmloverlay_dockwidget
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Control widget to configure remote sensing overlays.

    This file is part of mss.

    :copyright: Copyright 2017 Joern Ungermann
    :copyright: Copyright 2017-2018 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from builtins import str

from fs import open_fs
import logging
from lxml import etree, objectify
import os

# local application imports
from mslib.msui.mss_qt import QtGui, QtWidgets, get_open_filename
from mslib.msui.mss_qt import ui_kmloverlay_dockwidget as ui
from mslib.msui.mpl_map import KMLPatch
from mslib.utils import save_settings_pickle, load_settings_pickle


KMLPARSER = objectify.makeparser(strip_cdata=False)


class KMLOverlayControlWidget(QtWidgets.QWidget, ui.Ui_KMLOverlayDockWidget):
    """
    This class provides the interface for accessing KML files and
    adding the appropriate patches to the TopView canvas.
    """
    def __init__(self, parent=None, view=None):
        super(KMLOverlayControlWidget, self).__init__(parent)
        self.setupUi(self)
        self.view = view
        self.kml = None
        self.patch = None

        # Connect slots and signals.
        self.btSelectFile.clicked.connect(self.select_file)
        self.btLoadFile.clicked.connect(self.load_file)
        self.pbSelectColour.clicked.connect(self.select_colour)
        self.cbOverlay.stateChanged.connect(self.update_settings)
        self.dsbLineWidth.valueChanged.connect(self.update_settings)
        self.cbManualStyle.stateChanged.connect(self.update_settings)

        self.cbOverlay.setChecked(True)
        self.cbOverlay.setEnabled(False)
        self.cbManualStyle.setChecked(False)

        self.settings_tag = "kmldock"
        settings = load_settings_pickle(
            self.settings_tag, {"filename": "", "linewidth": 1, "colour": (0, 0, 0, 1)})

        self.leFile.setText(settings["filename"])
        self.dsbLineWidth.setValue(settings["linewidth"])

        palette = QtGui.QPalette(self.pbSelectColour.palette())
        colour = QtGui.QColor()
        colour.setRgbF(*settings["colour"])
        palette.setColor(QtGui.QPalette.Button, colour)
        self.pbSelectColour.setPalette(palette)

    def __del__(self):
        settings = {
            "filename": str(self.leFile.text()),
            "linewidth": self.dsbLineWidth.value(),
            "colour": self.get_color()
        }
        save_settings_pickle(self.settings_tag, settings)

    def get_color(self):
        button = self.pbSelectColour
        return QtGui.QPalette(button.palette()).color(QtGui.QPalette.Button).getRgbF()

    def update_settings(self):
        """
        Called when the visibility checkbox is toggled and hides/shows
        the overlay if loaded.
        """
        if self.view is not None and self.cbOverlay.isChecked() and self.patch is not None:
            self.view.plot_kml(self.patch)
            self.patch.update(self.cbManualStyle.isChecked(), self.get_color(), self.dsbLineWidth.value())
        elif self.patch is not None:
            self.view.plot_kml(None)

    def select_colour(self):
        button = self.pbSelectColour

        palette = QtGui.QPalette(button.palette())
        colour = palette.color(QtGui.QPalette.Button)
        colour = QtWidgets.QColorDialog.getColor(colour)
        if colour.isValid():
            palette.setColor(QtGui.QPalette.Button, colour)
            button.setPalette(palette)
        self.update_settings()

    def select_file(self):
        """Slot that opens a file dialog to choose a kml file
        """
        filename = get_open_filename(
            self, "Open KML Polygonal File", os.path.dirname(str(self.leFile.text())), "KML Files (*.kml)")
        if not filename:
            return
        self.leFile.setText(filename)

    def load_file(self):
        """
        Loads an KML file selected by the leFile box and constructs the
        corresponding patch.
        """
        _dirname, _name = os.path.split(self.leFile.text())
        _fs = open_fs(_dirname)
        if self.patch is not None:
            self.patch.remove()
            self.view.plot_kml(None)
            self.patch = None
            self.cbOverlay.setEnabled(False)
        try:
            with _fs.open(_name, 'r') as kmlf:
                self.kml = objectify.parse(kmlf, parser=KMLPARSER).getroot()
                self.patch = KMLPatch(self.view.map, self.kml,
                                      self.cbManualStyle.isChecked(), self.get_color(), self.dsbLineWidth.value())
            self.cbOverlay.setEnabled(True)
            if self.view is not None and self.cbOverlay.isChecked():
                self.view.plot_kml(self.patch)
        except (IOError, etree.XMLSyntaxError) as ex:
            logging.error("KML Overlay - %s: %s", type(ex), ex)
            QtWidgets.QMessageBox.critical(
                self, self.tr("KML Overlay"), self.tr(u"ERROR:\n{}\n{}".format(type(ex), ex)))



# -*- coding: utf-8 -*-
"""

    mslib.mswms._tests.test_wms
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to tests mswms.wms

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2017 Joern Ungermann
    :copyright: Copyright 2016-2018 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from past.builtins import basestring

import mslib.mswms.wms as wms


class Test_WMS(object):
    def test_get_capabilities(self):
        xml = wms.app.get_capabilities("http://localhost:8082")
        assert isinstance(xml, basestring), xml

    def test_produce_hsec_plot(self):
        environ = {
            'REQUEST_METHOD': 'GET', 'PATH_INFO': '/', 'SERVER_PROTOCOL': 'HTTP/1.1', 'HTTP_HOST': 'localhost:8081',
            'QUERY_STRING':
                'layers=ecmwf_EUR_LL015.PLDiv01&styles=&elevation=200&srs=EPSG%3A4326&format=image%2Fpng&'
                'request=GetMap&bgcolor=0xFFFFFF&height=376&dim_init_time=2012-10-17T12%3A00%3A00Z&width=479&'
                'version=1.1.1&bbox=-50.0%2C20.0%2C20.0%2C75.0&time=2012-10-17T12%3A00%3A00Z&'
                'exceptions=application%2Fvnd.ogc.se_xml&transparent=FALSE'}
        result = wms.app.produce_plot(environ, 'GetMap')
        assert len(result) == 2, result
        assert result[0] is not None, result
        assert result[1] == "image/png", result

    def test_produce_hsec_service_exception(self):
        environ = {
            'REQUEST_METHOD': 'GET', 'PATH_INFO': '/', 'SERVER_PROTOCOL': 'HTTP/1.1', 'HTTP_HOST': 'localhost:8081',
            'QUERY_STRING':
                'layers=ecmwf_EUR_LL015.PLDiv01&styles=&elevation=20&srs=EPSG%3A4326&format=image%2Fpng&'
                'request=GotMap&bgcolor=0xFFFFFF&height=376&dim_init_time=2012-10-17T12%3A00%3A00Z&width=479&'
                'version=1.1.1&bbox=-50.0%2C20.0%2C20.0%2C75.0&time=2012-10-17T12%3A00%3A00Z&'
                'exceptions=application%2Fvnd.ogc.se_xml&transparent=FALSE'}
        query_string = environ["QUERY_STRING"]
        for orig, fake in [
                ("dim_init_time=2012-10-17T12%3A00%3A00Z", "dim_init_time=20121017T12%3A00%3A00Z"),
                ("time=2012-10-17T12%3A00%3A00Z", "time=201210-17T12%3A00%3A00Z"),
                ("time=2012-10-17T12%3A00%3A00Z", "time=2012-01-17T12%3A00%3A00Z"),
                ("&dim_init_time=2012-10-17T12%3A00%3A00Z", ""),
                ("&time=2012-10-17T12%3A00%3A00Z", ""),
                ("srs=EPSG%3A4326", "srs=EPSH%3A4326"),
                ("srs=EPSG%3A4326", "srs=EPSG%3AABCD"),
                ("srs=EPSG%3A4326", "srs=EPSG%3A6666"),
                ("ecmwf_EUR_LL015.PLDiv01", "PLDiv01"),
                ("ecmwf_EUR_LL015.PLDiv01", "ecmwf_EUR_LL015.PLDav01"),
                ("ecmwf_EUR_LL015.PLDiv01", "ecmwf_AUR_LL015.PLDiv01"),
                ("format=image%2Fpng", "format=omage%2Fpng"),
                ("bbox=-50.0%2C20.0%2C20.0%2C75.0", "bbox=-abcd%2C20.0%2C20.0%2C75.0")]:
            environ["QUERY_STRING"] = query_string.replace(orig, fake)
            result = wms.app.produce_plot(environ, 'GetMap')
            assert len(result) == 2, result
            assert isinstance(result[0], basestring), result
            assert result[0].count("ServiceExceptionReport") > 0, result
            assert result[1] == "text/xml", result

    def test_produce_vsec_plot(self):
        environ = {
            'REQUEST_METHOD': 'GET', 'PATH_INFO': '/', 'SERVER_PROTOCOL': 'HTTP/1.1', 'HTTP_HOST': 'localhost:8081',
            'QUERY_STRING':
                'layers=ecmwf_EUR_LL015.VS_HV01&styles=&srs=VERT%3ALOGP&format=image%2Fpng&'
                'request=GetMap&bgcolor=0xFFFFFF&height=245&dim_init_time=2012-10-17T12%3A00%3A00Z&width=842&'
                'version=1.1.1&bbox=201%2C500.0%2C10%2C100.0&time=2012-10-17T12%3A00%3A00Z&'
                'exceptions=application%2Fvnd.ogc.se_xml&path=52.78%2C-8.93%2C48.08%2C11.28&transparent=FALSE'}
        result = wms.app.produce_plot(environ, 'GetMap')
        assert len(result) == 2, result
        assert result[0] is not None, result
        assert result[1] == "image/png", result

    def test_produce_vsec_service_exception(self):
        environ = {
            'REQUEST_METHOD': 'GET', 'PATH_INFO': '/', 'SERVER_PROTOCOL': 'HTTP/1.1', 'HTTP_HOST': 'localhost:8081',
            'QUERY_STRING':
                'layers=ecmwf_EUR_LL015.VS_HV01&styles=&srs=VERT%3ALOGP&format=image%2Fpng&'
                'request=GetMap&bgcolor=0xFFFFFF&height=245&dim_init_time=2012-10-17T12%3A00%3A00Z&width=842&'
                'version=1.1.1&bbox=201%2C500.0%2C10%2C100.0&time=2012-10-17T12%3A00%3A00Z&'
                'exceptions=application%2Fvnd.ogc.se_xml&path=52.78%2C-8.93%2C48.08%2C11.28&transparent=FALSE'}
        query_string = environ["QUERY_STRING"]
        for orig, fake in [
                ("time=2012-10-17T12%3A00%3A00Z", "time=2012-01-17T12%3A00%3A00Z"),
                ("&dim_init_time=2012-10-17T12%3A00%3A00Z", ""),
                ("&time=2012-10-17T12%3A00%3A00Z", ""),
                ("layers=ecmwf_EUR_LL015.VS_HV01", "layers=ecmwf_AUR_LL015.VS_HV01"),
                ("layers=ecmwf_EUR_LL015.VS_HV01", "layers=ecmwf_EUR_LL015.VS_HV99"),
                ("format=image%2Fpng", "format=omage%2Fpng"),
                ("path=52.78%2C-8.93%2C48.08%2C11.28", "path=aaaa%2C-8.93%2C48.08%2C11.28"),
                ("&path=52.78%2C-8.93%2C48.08%2C11.28", ""),
                ("bbox=201%2C500.0%2C10%2C100.0", "bbox=aaa%2C500.0%2C10%2C100.0")]:
            environ["QUERY_STRING"] = query_string.replace(orig, fake)
            result = wms.app.produce_plot(environ, 'GetMap')
            assert len(result) == 2, result
            assert isinstance(result[0], basestring), result
            assert result[0].count("ServiceExceptionReport") > 0, result
            assert result[1] == "text/xml", result



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_kmloverlay_dockwidget.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_KMLOverlayDockWidget(object):
    def setupUi(self, KMLOverlayDockWidget):
        KMLOverlayDockWidget.setObjectName("KMLOverlayDockWidget")
        KMLOverlayDockWidget.resize(649, 120)
        self.verticalLayout = QtWidgets.QVBoxLayout(KMLOverlayDockWidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.cbOverlay = QtWidgets.QCheckBox(KMLOverlayDockWidget)
        self.cbOverlay.setObjectName("cbOverlay")
        self.horizontalLayout_5.addWidget(self.cbOverlay)
        self.leFile = QtWidgets.QLineEdit(KMLOverlayDockWidget)
        self.leFile.setObjectName("leFile")
        self.horizontalLayout_5.addWidget(self.leFile)
        self.btSelectFile = QtWidgets.QToolButton(KMLOverlayDockWidget)
        self.btSelectFile.setObjectName("btSelectFile")
        self.horizontalLayout_5.addWidget(self.btSelectFile)
        self.btLoadFile = QtWidgets.QToolButton(KMLOverlayDockWidget)
        self.btLoadFile.setObjectName("btLoadFile")
        self.horizontalLayout_5.addWidget(self.btLoadFile)
        self.verticalLayout.addLayout(self.horizontalLayout_5)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.cbManualStyle = QtWidgets.QCheckBox(KMLOverlayDockWidget)
        self.cbManualStyle.setObjectName("cbManualStyle")
        self.horizontalLayout_7.addWidget(self.cbManualStyle)
        self.pbSelectColour = QtWidgets.QPushButton(KMLOverlayDockWidget)
        self.pbSelectColour.setObjectName("pbSelectColour")
        self.horizontalLayout_7.addWidget(self.pbSelectColour)
        self.label_2 = QtWidgets.QLabel(KMLOverlayDockWidget)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_7.addWidget(self.label_2)
        self.dsbLineWidth = QtWidgets.QDoubleSpinBox(KMLOverlayDockWidget)
        self.dsbLineWidth.setSingleStep(0.5)
        self.dsbLineWidth.setProperty("value", 1.0)
        self.dsbLineWidth.setObjectName("dsbLineWidth")
        self.horizontalLayout_7.addWidget(self.dsbLineWidth)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem)
        self.verticalLayout.addLayout(self.horizontalLayout_7)
        self.label = QtWidgets.QLabel(KMLOverlayDockWidget)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem1)

        self.retranslateUi(KMLOverlayDockWidget)
        QtCore.QMetaObject.connectSlotsByName(KMLOverlayDockWidget)

    def retranslateUi(self, KMLOverlayDockWidget):
        _translate = QtCore.QCoreApplication.translate
        KMLOverlayDockWidget.setWindowTitle(_translate("KMLOverlayDockWidget", "KML Overlay"))
        self.cbOverlay.setToolTip(_translate("KMLOverlayDockWidget", "Hide/unhide loaded KML overlay."))
        self.cbOverlay.setText(_translate("KMLOverlayDockWidget", "KML Overlay"))
        self.btSelectFile.setToolTip(_translate("KMLOverlayDockWidget", "Specify KML file using a file dialog."))
        self.btSelectFile.setText(_translate("KMLOverlayDockWidget", "..."))
        self.btLoadFile.setToolTip(_translate("KMLOverlayDockWidget", "Load specified KML file."))
        self.btLoadFile.setText(_translate("KMLOverlayDockWidget", "load"))
        self.cbManualStyle.setToolTip(_translate("KMLOverlayDockWidget", "Overwrite styles of KML document with colour/linewidth specified to the right."))
        self.cbManualStyle.setText(_translate("KMLOverlayDockWidget", "manual style"))
        self.pbSelectColour.setToolTip(_translate("KMLOverlayDockWidget", "Colour for KML overlay."))
        self.pbSelectColour.setText(_translate("KMLOverlayDockWidget", "Colour"))
        self.label_2.setToolTip(_translate("KMLOverlayDockWidget", "Line width for KML overlay."))
        self.label_2.setText(_translate("KMLOverlayDockWidget", "line width"))
        self.label.setToolTip(_translate("KMLOverlayDockWidget", "KML support is experimental. Please communicate important missing features to the MSS issue tracker!"))
        self.label.setText(_translate("KMLOverlayDockWidget", "!Experimental Feature! Not all KML files and contained features will work."))




# -*- coding: utf-8 -*-
"""

    mslib.msui._tests.test_kmloverlay_dockwidget
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to tests msui.kmloverlay_dockwidget

    This file is part of mss.

    :copyright: Copyright 2017 Joern Ungermann
    :copyright: Copyright 2017-2018 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
import sys
import mock
from mslib.msui.mss_qt import QtWidgets, QtCore, QtTest
import mslib.msui.kmloverlay_dockwidget as kd


class Test_KmlOverlayDockWidget(object):
    def setup(self):
        self.application = QtWidgets.QApplication(sys.argv)
        self.view = mock.Mock()
        self.view.map = mock.Mock(side_effect=lambda x, y: (x, y))
        self.view.map.plot = mock.Mock(return_value=[mock.Mock()])
        self.window = kd.KMLOverlayControlWidget(view=self.view)
        self.window.show()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWaitForWindowExposed(self.window)
        # start load test
        QtWidgets.QApplication.processEvents()

    def teardown(self):
        QtWidgets.QApplication.processEvents()
        self.application.quit()
        QtWidgets.QApplication.processEvents()
        del self.window

    @mock.patch("mslib.msui.mss_qt.QtWidgets.QMessageBox")
    def test_load(self, mockbox):
        """
        Test for flawless loading of parsing of KML files
        """
        for sample in ["folder.kml", "line.kml"]:
            path = os.path.join(os.path.dirname(__file__), "..", "..", "..", "docs", "samples", "kml", sample)
            self.window.leFile.setText(path)
            QtTest.QTest.mouseClick(self.window.btLoadFile, QtCore.Qt.LeftButton)
            QtWidgets.QApplication.processEvents()
            assert mockbox.critical.call_count == 0

    def test_styles(self):
        """
        Test for changing styles and toggling visibility
        """
        self.test_load()
        # disable/enable overlay
        QtTest.QTest.mouseClick(self.window.cbOverlay, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.mouseClick(self.window.cbOverlay, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        # disable/enable styles
        QtTest.QTest.mouseClick(self.window.cbManualStyle, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.mouseClick(self.window.cbManualStyle, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()

    @mock.patch("mslib.msui.mss_qt.QtWidgets.QMessageBox")
    def test_load_error(self, mockbox):
        """
        Test that program mitigates loading a non-existing file
        """
        # load a non existing path
        path = os.path.join(os.path.dirname(__file__), "..", "..", "..", "docs",
                            "samples", "satellite_tracks", "satellite_predictor.txt")
        self.window.leFile.setText(path)
        QtTest.QTest.mouseClick(self.window.btLoadFile, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        assert mockbox.critical.call_count == 1



# -*- coding: utf-8 -*-
"""

    mslib.msui.mpl_map
    ~~~~~~~~~~~~~~~~~~

    Map canvas for the top view.
    As Matplotlib's Basemap is primarily designed to produce static plots,
    we derived a class MapCanvas to allow for a certain degree of
    interactivity. MapCanvas extends Basemap by functionality to, for
    instance, automatically draw a graticule. It also keeps references to
    plotted map elements to allow the user to toggle their visibility, or
    to redraw when map section (zoom/pan) or projection have been changed.

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2018 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import division

from builtins import str

import logging

# related third party imports
import numpy as np
import matplotlib
import matplotlib.path as mpath
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from matplotlib import patheffects
import mpl_toolkits.basemap as basemap
try:
    import mpl_toolkits.basemap.pyproj as pyproj
except ImportError:
    import pyproj

from mslib.msui import mpl_pathinteractor as mpl_pi
from mslib.msui import trajectory_item_tree as titree


class MapCanvas(basemap.Basemap):
    """Derivative of mpl_toolkits.basemap, providing additional methods to
       automatically draw a graticule and to redraw specific map elements.
    """

    def __init__(self, identifier=None, CRS=None, BBOX_UNITS=None,
                 traj_item_tree=None, appearance=None, **kwargs):
        """New constructor automatically adds coastlines, continents, and
           a graticule to the map.

        Keyword arguments are the same as for mpl_toolkits.basemap.

        Additional arguments:
        CRS -- string describing the coordinate reference system of the map.
        BBOX_UNITS -- string describing the units of the map coordinates.

        """
        # Coordinate reference system identifier and coordinate system units.
        self.crs = CRS if CRS is not None else self.crs if hasattr(self, "crs") else None
        self.bbox_units = BBOX_UNITS if BBOX_UNITS else self.bbox_units \
            if hasattr(self, "bbox_units") else None

        # Dictionary containing map appearance settings.
        param_appearance = appearance if appearance else self.appearance \
            if hasattr(self, "appearance") else {}
        default_appearance = {"draw_graticule": True,
                              "draw_coastlines": True,
                              "fill_waterbodies": True,
                              "fill_continents": True,
                              "colour_water": ((153 / 255.), (255 / 255.), (255 / 255.), (255 / 255.)),
                              "colour_land": ((204 / 255.), (153 / 255.), (102 / 255.), (255 / 255.))}
        default_appearance.update(param_appearance)
        self.appearance = default_appearance

        # Identifier of this map canvas (used to query data structures that
        # are observed by different views).
        self.identifier = identifier if identifier else self.identifier \
            if hasattr(self, "identifier") else None

        # Call the Basemap constructor. If a cylindrical projection was used
        # before, Basemap stores an EPSG code that will not be changed if
        # the Basemap constructor is called a second time. Hence, we have to
        # delete the attribute (mr, 08Feb2013).
        if hasattr(self, "epsg"):
            del self.epsg
        super(MapCanvas, self).__init__(**kwargs)
        self.kwargs = kwargs

        # Set up the map appearance.
        if self.appearance["draw_coastlines"]:
            self.map_coastlines = self.drawcoastlines(zorder=3)
            self.map_countries = self.drawcountries(zorder=3)
        else:
            self.map_coastlines = None
            self.map_countries = None
        if self.appearance["fill_waterbodies"]:
            self.map_boundary = self.drawmapboundary(fill_color=self.appearance["colour_water"])
        else:
            self.map_boundary = None

        # zorder = 0 is necessary to paint over the filled continents with
        # scatter() for drawing the flight tracks and trajectories.
        # Curiously, plot() works fine without this setting, but scatter()
        # doesn't.
        if self.appearance["fill_continents"]:
            self.map_continents = self.fillcontinents(color=self.appearance["colour_land"],
                                                      lake_color=self.appearance["colour_water"],
                                                      zorder=1)
        else:
            self.map_continents = None

        self.image = None

        # Print CRS identifier into the figure.
        if self.crs is not None:
            if hasattr(self, "crs_text"):
                self.crs_text.set_text(self.crs)
            else:
                self.crs_text = self.ax.figure.text(0, 0, self.crs)

        if self.appearance["draw_graticule"]:
            try:
                self._draw_auto_graticule()
            except Exception as ex:
                logging.error(u"ERROR: cannot plot graticule (message: {} - '{}')".format(type(ex), ex))
        else:
            self.map_parallels = None
            self.map_meridians = None

        # self.warpimage() # disable fillcontinents when loading bluemarble
        self.ax.set_autoscale_on(False)

        # Connect to the trajectory item tree, if defined.
        self.traj_item_tree = traj_item_tree if traj_item_tree is not None else self.traj_item_tree if hasattr(
            self, "traj_item_tree") else None
        if traj_item_tree is not None:
            self.set_trajectory_tree(traj_item_tree)

        # The View may be destroyed and then this class is left dangling due to the connected
        # trajectories unless we disconnect it.
        self.ax.figure.canvas.destroyed.connect(self.disconnectTrajectories)

    def set_identifier(self, identifier):
        self.identifier = identifier

    def set_axes_limits(self, ax=None):
        """See Basemap.set_axes_limits() for documentation.

        This function is overridden in MapCanvas as a workaround to a problem
        in Basemap.set_axes_limits() that occurs in interactive matplotlib
        mode. If matplotlib.is_interactive() is True, Basemap.set_axes_limits()
        tries to redraw the canvas by accessing the pylab figure manager.
        If the matplotlib object is embedded in a Qt application, this manager
        is not available and an exception is raised. Hence, the interactive
        mode is disabled here before the original Basemap.set_axes_limits()
        method is called. It is restored afterwards.
        """
        intact = matplotlib.is_interactive()
        matplotlib.interactive(False)
        super(MapCanvas, self).set_axes_limits(ax=ax)
        matplotlib.interactive(intact)

    def _draw_auto_graticule(self):
        """Draw an automatically spaced graticule on the map.
        """
        # Compute some map coordinates that are required below for the automatic
        # determination of which meridians and parallels to draw.
        axis = self.ax.axis()
        upperLeftCornerLon, upperLeftCornerLat = self.__call__(
            axis[0], axis[3], inverse=True)
        lowerRightCornerLon, lowerRightCornerLat = self.__call__(
            axis[1], axis[2], inverse=True)
        middleUpperBoundaryLon, middleUpperBoundaryLat = self.__call__(
            np.mean([axis[0], axis[1]]), axis[3], inverse=True)
        middleLowerBoundaryLon, middleLowerBoundaryLat = self.__call__(
            np.mean([axis[0], axis[1]]), axis[2], inverse=True)

        # Determine which parallels and meridians should be drawn.
        #   a) determine which are the minimum and maximum visible
        #      longitudes and latitudes, respectively. These
        #      values depend on the map projection.
        if self.projection in ['stere', 'lcc']:
            # For stereographic projections: Draw meridians from the minimum
            # longitude contained in the map at one of the four corners to the
            # maximum longitude at one of these corner points. If
            # the southern or norther pole  is contained in the map, draw all
            # meridians around the globe.
            # Draw parallels from the min latitude contained in the map at
            # one of the four corners OR the middle top or bottom to the
            # maximum latitude at one of these six points.
            # If the map centre in contained in the map, replace either
            # start or end latitude by the centre latitude (whichever is
            # smaller/larger).

            # check if centre point of projection is contained in the map,
            # use projection coordinates for this test
            centre_x = self.projparams["x_0"]
            centre_y = self.projparams["y_0"]
            centre_lon, centre_lat = self.__call__(centre_x, centre_y, inverse=True)
            if centre_lat > 0:
                pole_lon, pole_lat = 0, 90
            else:
                pole_lon, pole_lat = 0, -90
            pole_x, pole_y = self.__call__(pole_lon, pole_lat)
            if self.urcrnrx > self.llcrnrx:
                contains_pole = (self.llcrnrx <= pole_x <= self.urcrnrx)
            else:
                contains_pole = (self.llcrnrx >= pole_x >= self.urcrnrx)
            if self.urcrnry > self.llcrnry:
                contains_pole = contains_pole and (self.llcrnry <= pole_y <= self.urcrnry)
            else:
                contains_pole = contains_pole and (self.llcrnry >= pole_y >= self.urcrnry)

            # merdidians
            if contains_pole:
                mapLonStart = -180.
                mapLonStop = 180.
            else:
                mapLonStart = min(upperLeftCornerLon, self.llcrnrlon,
                                  self.urcrnrlon, lowerRightCornerLon)
                mapLonStop = max(upperLeftCornerLon, self.llcrnrlon,
                                 self.urcrnrlon, lowerRightCornerLon)
            # parallels
            mapLatStart = min(middleLowerBoundaryLat, lowerRightCornerLat,
                              self.llcrnrlat,
                              middleUpperBoundaryLat, upperLeftCornerLat,
                              self.urcrnrlat)
            mapLatStop = max(middleLowerBoundaryLat, lowerRightCornerLat,
                             self.llcrnrlat,
                             middleUpperBoundaryLat, upperLeftCornerLat,
                             self.urcrnrlat)
            if contains_pole:
                centre_lat = self.projparams["lat_0"]
                mapLatStart = min(mapLatStart, centre_lat)
                mapLatStop = max(mapLatStop, centre_lat)
        else:
            # for other projections (preliminary): difference between the
            # lower left and the upper right corner.
            mapLonStart = self.llcrnrlon
            mapLonStop = self.urcrnrlon
            mapLatStart = self.llcrnrlat
            mapLatStop = self.urcrnrlat

        # b) parallels and meridians can be drawn with a spacing of
        #      >spacingValues< degrees. Determine the appropriate
        #      spacing for the lon/lat differences: about 10 lines
        #      should be drawn in each direction. (The following lines
        #      filter the spacingValues list for all spacing values
        #      that are larger than lat/lon difference / 10, then
        #      take the first value (first values that's larger)).
        spacingValues = [0.05, 0.1, 0.25, 0.5, 1, 2, 5, 10, 20, 30, 40]
        deltaLon = mapLonStop - mapLonStart
        deltaLat = mapLatStop - mapLatStart
        spacingLon = [i for i in spacingValues if i > (deltaLon / 11.)][0]
        spacingLat = [i for i in spacingValues if i > (deltaLat / 11.)][0]

        #   c) parallels and meridians start at the first value in the
        #      spacingLon/Lat grid that's smaller than the lon/lat of the
        #      lower left corner; they stop at the first values in the
        #      grid that's larger than the lon/lat of the upper right corner.
        lonStart = np.floor((mapLonStart / spacingLon)) * spacingLon
        lonStop = np.ceil((mapLonStop / spacingLon)) * spacingLon
        latStart = np.floor((mapLatStart / spacingLat)) * spacingLat
        latStop = np.ceil((mapLatStop / spacingLat)) * spacingLat

        #   d) call the basemap methods to draw the lines in the determined
        #      range.
        self.map_parallels = self.drawparallels(np.arange(latStart, latStop,
                                                          spacingLat),
                                                labels=[1, 1, 0, 0], zorder=3)
        self.map_meridians = self.drawmeridians(np.arange(lonStart, lonStop,
                                                          spacingLon),
                                                labels=[0, 0, 0, 1], zorder=3)

    def set_graticule_visible(self, visible=True):
        """Set the visibily of the graticule.

        Removes a currently visible graticule by deleting internally stored
        line and text objects representing graticule lines and labels, then
        redrawing the map canvas.

        See http://www.mail-archive.com/matplotlib-users@lists.sourceforge.net/msg09349.html
        """
        self.appearance["draw_graticule"] = visible
        if visible and self.map_parallels is None and self.map_meridians is None:
            # Draw new graticule if visible is True and no current graticule
            # exists.
            self._draw_auto_graticule()
            # Update the figure canvas.
            self.ax.figure.canvas.draw()
        elif not visible and self.map_parallels is not None and self.map_meridians is not None:
            # If visible if False, remove current graticule if one exists.
            # Every item in self.map_parallels and self.map_meridians is
            # a tuple of a list of lines and a list of text labels.
            for item in self.map_parallels.values():
                for line in item[0]:
                    line.remove()
                for text in item[1]:
                    text.remove()
            for item in self.map_meridians.values():
                for line in item[0]:
                    line.remove()
                for text in item[1]:
                    text.remove()
            self.map_parallels = None
            self.map_meridians = None
            # Update the figure canvas.
            self.ax.figure.canvas.draw()

    def set_fillcontinents_visible(self, visible=True, land_color=None,
                                   lake_color=None):
        """Set the visibility of continent fillings.
        """
        if land_color is not None:
            self.appearance["colour_land"] = land_color
        if lake_color is not None:
            self.appearance["colour_water"] = lake_color
        self.appearance["fill_continents"] = visible

        if visible and self.map_continents is None:
            # zorder = 0 is necessary to paint over the filled continents with
            # scatter() for drawing the flight tracks and trajectories.
            # Curiously, plot() works fine without this setting, but scatter()
            # doesn't.
            self.map_continents = self.fillcontinents(color=self.appearance["colour_land"],
                                                      lake_color=self.appearance["colour_water"],
                                                      zorder=1)
            self.ax.figure.canvas.draw()
        elif not visible and self.map_continents is not None:
            # Remove current fills. They are stored as a list of polygon patches
            # in self.map_continents.
            for patch in self.map_continents:
                patch.remove()
            self.map_continents = None
            self.ax.figure.canvas.draw()
        elif visible and self.map_continents is not None:
            # Colours have changed: Remove the old fill and redraw.
            for patch in self.map_continents:
                patch.remove()
            self.map_continents = self.fillcontinents(color=self.appearance["colour_land"],
                                                      lake_color=self.appearance["colour_water"],
                                                      zorder=1)
            self.ax.figure.canvas.draw()

    def set_coastlines_visible(self, visible=True):
        """Set the visibility of coastlines and country borders.
        """
        self.appearance["draw_coastlines"] = visible
        if visible and self.map_coastlines is None and self.map_countries is None:
            self.map_coastlines = self.drawcoastlines(zorder=3)
            self.map_countries = self.drawcountries(zorder=3)
            self.ax.figure.canvas.draw()
        elif not visible and self.map_coastlines is not None and self.map_countries is not None:
            self.map_coastlines.remove()
            self.map_countries.remove()
            del self.cntrysegs
            self.map_coastlines = None
            self.map_countries = None
            self.ax.figure.canvas.draw()

    def set_mapboundary_visible(self, visible=True, bg_color='#99ffff'):
        """
        """
        # TODO: This doesn't work. Removing the map background only makes sense
        #      if there's a second image underneath this map. Maybe we should work
        #      with alpha values instead.
        self.appearance["fill_waterbodies"] = visible
        self.appearance["colour_water"] = bg_color

        if not visible and self.map_boundary is not None:
            self.map_boundary.remove()
            self.map_boundary = None
            self.ax.figure.canvas.draw()
        elif visible:
            self.map_boundary = self.drawmapboundary(fill_color=bg_color)
            self.ax.figure.canvas.draw()

    def update_with_coordinate_change(self, kwargs_update=None):
        """Redraws the entire map. This is necessary after zoom/pan operations.

        Determines corner coordinates of the current axes, removes all items
        belonging the the current map and draws a new one by calling
        self.__init__().

        DRAWBACK of this approach is that the map coordinate system changes, as
        basemap always takes the lower left axis corner as (0,0). This means
        that all other objects on the matplotlib canvas (flight track, markers,
        ..) will be incorrectly placed after a redraw. Their coordinates need
        to be adjusted by 1) transforming their coordinates to lat/lon BEFORE
        the map is redrawn, 2) redrawing the map, 3) transforming the stored
        lat/lon coordinates to the new map coordinates.
        """
        # Convert the current axis corners to lat/lon coordinates.
        axis = self.ax.axis()
        self.kwargs['llcrnrlon'], self.kwargs['llcrnrlat'] = \
            self.__call__(axis[0], axis[2], inverse=True)
        self.kwargs['urcrnrlon'], self.kwargs['urcrnrlat'] = \
            self.__call__(axis[1], axis[3], inverse=True)

        logging.debug("corner coordinates (lat/lon): ll(%.2f,%.2f), ur(%.2f,%.2f)",
                      self.kwargs['llcrnrlat'], self.kwargs['llcrnrlon'],
                      self.kwargs['urcrnrlat'], self.kwargs['urcrnrlon'])

        if self.kwargs["projection"] in ["cyl"]:
            # Latitudes in cylindrical projection need to be within -90..90.
            self.kwargs['llcrnrlat'] = max(self.kwargs['llcrnrlat'], -90)
            self.kwargs['urcrnrlat'] = max(self.kwargs['urcrnrlat'], -89)
            self.kwargs['llcrnrlat'] = min(self.kwargs['llcrnrlat'], 89)
            self.kwargs['urcrnrlat'] = min(self.kwargs['urcrnrlat'], 90)
            # Longitudes in cylindrical projection need to be within -360..540.
            self.kwargs["llcrnrlon"] = max(self.kwargs["llcrnrlon"], -360)
            self.kwargs["urcrnrlon"] = max(self.kwargs["urcrnrlon"], -359)
            self.kwargs["llcrnrlon"] = min(self.kwargs["llcrnrlon"], 539)
            self.kwargs["urcrnrlon"] = min(self.kwargs["urcrnrlon"], 540)

        # Remove the current map artists.
        grat_vis = self.appearance["draw_graticule"]
        self.set_graticule_visible(False)
        self.appearance["draw_graticule"] = grat_vis
        if self.map_coastlines is not None:
            self.map_coastlines.remove()

        if self.image is not None:
            self.image.remove()
            self.image = None

        # Refer to Basemap.drawcountries() on how to remove country borders.
        # In addition to the matplotlib lines, the loaded country segment data
        # needs to be loaded. THE SAME NEEDS TO BE DONE WITH RIVERS ETC.
        if self.map_countries is not None:
            self.map_countries.remove()
            del self.cntrysegs

        # map_boundary is None for rectangular projections (basemap simply sets
        # the backgorund colour).
        if self.map_boundary is not None:
            try:
                # TODO: review
                # FIX (mr, 15Oct2012) -- something seems to have changed in
                # newer Matplotlib versions: this causes an exception when
                # the user zooms/pans..
                self.map_boundary.remove()
            except NotImplementedError:
                pass

        cont_vis = self.appearance["fill_continents"]
        self.set_fillcontinents_visible(False)
        self.appearance["fill_continents"] = cont_vis

        # POSSIBILITY A): Call self.__init__ again with stored keywords.
        # Update kwargs if new parameters such as the map region have been
        # given.
        if kwargs_update:
            self.kwargs.update(kwargs_update)
        self.__init__(**self.kwargs)

        self.update_trajectory_items()

        # TODO: HOW TO MAKE THIS MORE EFFICIENT.
        # POSSIBILITY B): Only set the Basemap parameters that influence the
        # plot (corner lat/lon, x/y, width/height, ..) and re-define the
        # polygons that represent the coastlines etc. In Basemap, they are
        # defined in __init__(), at the very bottom (the code that comes
        # with the comments
        #    >> read in coastline polygons, only keeping those that
        #    >> intersect map boundary polygon.
        # ). Basemap only loads the coastline data that is actually displayed.
        # If we only change llcrnrlon etc. here and replot coastlines etc.,
        # the polygons and the map extent will remain the same.
        # However, it should be possible to make a map change WITHOUT changing
        # coordinates.
        #

    # self.llcrnrlon = llcrnrlon
    # self.llcrnrlat = llcrnrlat
    # self.urcrnrlon = urcrnrlon
    # self.urcrnrlat = urcrnrlat
    # self.llcrnrx = axis[0]
    # self.llcrnry = axis[2]
    # self.urcrnrx = axis[1]
    # self.urcrnry = axis[3]

    def imshow(self, X, **kwargs):
        """Overloads basemap.imshow(). Deletes any existing image and
           redraws the figure after the new image has been plotted.
        """
        if self.image is not None:
            self.image.remove()
        self.image = super(MapCanvas, self).imshow(X, zorder=2, **kwargs)
        self.ax.figure.canvas.draw()
        return self.image

    def set_trajectory_tree(self, tree):
        """Set a reference to the tree data structure containing the information
           about the elements plotted on the map (e.g. flight tracks,
           trajectories).
        """
        logging.debug("registering trajectory tree model")
        # Disconnect old tree, if defined.
        if self.traj_item_tree is not None:
            self.traj_item_tree.dataChanged.disconnect(self.update_from_trajectory_tree)
        # Set and connect new tree.
        self.traj_item_tree = tree
        self.traj_item_tree.dataChanged.connect(self.update_from_trajectory_tree)
        # Draw tree items.
        self.update_trajectory_items()

    def disconnectTrajectories(self):
        if self.traj_item_tree is not None:
            self.traj_item_tree.dataChanged.disconnect(self.update_from_trajectory_tree)

    def update_from_trajectory_tree(self, index1, index2):
        """This method should be connected to the 'dataChanged()' signal
           of the MapItemsTree.

        NOTE: If index1 != index2, the entire tree will be updated (I haven't
              found out so far how to get all the items between the two
              indices).
        """
        if self.traj_item_tree is None:
            return
        # Update the map elements if the change that occured in traj_item_tree
        # affected a LagrantoMapItem.
        item = index1.internalPointer()
        item2 = index2.internalPointer()
        if isinstance(item, titree.LagrantoMapItem):
            last_change = self.traj_item_tree.getLastChange()
            # Update the given item or the entire tree if the two given
            # items are different.
            self.update_trajectory_items(item=item if item == item2 else None,
                                         mode=last_change)

    def update_trajectory_items(self, item=None, mode="DRAW_EVERYTHING"):
        """Draw or update map elements.

        The map items tree is traversed downward starting at the item
        given as the argument 'item' (i.e. the given item and all its
        children will be processed). Also, the path from the given
        item to the root of the tree will be followed to check if any
        property (visibility, colour) is inherited from a higher level.

        If no item is given, the traverse will start at the root, i.e.
        all items on the map will be drawn or updated.

        Keyword arguments:
        item -- LagrantoMapItem instance from which the traversal should
                be started.
        mode -- string with information on what should be done (draw vs.
                update):

                DRAW_EVERYTHING -- draw or redraw all elements on the map.
                FLIGHT_TRACK_ADDED -- draw the flight track represented
                    by 'item' (including markers).
                TRAJECTORY_DIR_ADDED -- draw the trajectory represented
                    by 'item' (including markers).
                MARKER_CHANGE -- markers for the current item (and all items in
                    its subtree) have been changed. Delete the old graphics
                    instances and draw new markers.
                GXPROPERTY_CHANGE -- a graphics property has been changed (e.g.
                    colour). Perform a pylab.setp() call to change the property.
                VISIBILITY_CHANGE -- same action as for GXPROPERTY_CHANGE
        """
        # If no trajectory item tree is defined exit the method.
        if self.traj_item_tree is None:
            return

        logging.debug("updating trajectory items on map <%s>, mode %s",
                      self.identifier, mode)

        # If no item is given start at the root of the 'traj_item_tree'.
        if item is None:
            item = self.traj_item_tree.getRootItem()
            logging.debug("processing all trajectory items")

        # This method can only operate on LagrantoMapItems -- raise an exception
        # if the argument is not of this type.
        elif not isinstance(item, titree.LagrantoMapItem):
            raise TypeError("This method can only process 'LagrantoMapItem's.")

        # Traverse the tree upward from the current item to the tree root,
        # to check if there are any properties that are inherited from an
        # item node at a higher level to the current item (e.g. an item
        # inherits visibility, colour and timeMarkerInterval from its
        # ancestors).
        #
        # Default values: The item is visible and inherits no other properties.
        parent_properties = {"visible": True,
                             "colour": None,
                             "linestyle": None,
                             "linewidth": None,
                             "timeMarkerInterval": None}
        parent = item.parent()  # this would be None if item == rootItem
        while parent is not None:  # loop until root has been reached
            # Set visible to False if one item along the path to the root is
            # found to be set to invisible.
            parent_properties["visible"] = parent.isVisible(self.identifier) and parent_properties["visible"]

            # Inherit colour and time marker interval of the highest tree level
            # on which they are set (set the current property to the value
            # encountered on the current level if the property is not None,
            # otherwise keep the old value -- since we're moving upward,
            # the property will eventually have the value of the highest
            # level on which is was not None).
            for propName in ["colour", "linestyle", "linewidth",
                             "timeMarkerInterval"]:
                prop = parent.getGxElementProperty("general", propName)
                parent_properties[propName] = prop if prop \
                    else parent_properties[propName]

            # Move up one level.
            parent = parent.parent()

        # Iterative tree traversal: Put the item and its attributes on a
        # stack.
        item_stack = [(item, parent_properties)]

        # Operate on the items in the stack until the stack is empty.
        while len(item_stack) > 0:

            # Get a new item from the stack.
            item, parent_properties = item_stack.pop()

            # Its visibility is determined by its own visibility status
            # and that of its parent (i.e. the subtree it is part of).
            item_properties = {}
            item_properties["visible"] = item.isVisible(self.identifier) and parent_properties["visible"]
            #
            for propName in ["colour", "linestyle", "linewidth",
                             "timeMarkerInterval"]:
                item_properties[propName] = parent_properties[propName] \
                    if parent_properties[propName] \
                    else item.getGxElementProperty("general",
                                                   propName)

            # Push all children of the current item onto the stack that are
            # instances of LagrantoMapItem (VariableItems are not plotted on
            # the map).
            item_stack.extend([(child, item_properties) for child in item.childItems
                               if isinstance(child, titree.LagrantoMapItem)])

            # Plotting and graphics property update operations can only be
            # performed for flight tracks and trajectories.
            if not (isinstance(item, titree.FlightTrackItem) or isinstance(item, titree.TrajectoryItem)):
                continue

            if mode in ["GXPROPERTY_CHANGE", "VISIBILITY_CHANGE"]:
                # Set the visibility of all graphics elements of the current item
                # to 'itemVisible'.
                # NOTE: We do not have to test if the item is of type
                #   FlightTrackItem or TrajectoryItem, since only these
                #   two classes contain fields other that 'general' in
                #   gxElements.
                for element in item.getGxElements():
                    if element == "general":
                        continue
                    elif element == "lineplot":
                        plt.setp(
                            item.getGxElementProperty(element,
                                                      u"instance::{}".format(self.identifier)),
                            visible=item_properties["visible"],
                            color=item_properties["colour"],
                            linestyle=item_properties["linestyle"],
                            linewidth=item_properties["linewidth"])
                    elif element == "timemarker":
                        plt.setp(
                            item.getGxElementProperty(element,
                                                      u"instance::{}".format(self.identifier)),
                            visible=item_properties["visible"],
                            color=item_properties["colour"])

            elif mode in ["FLIGHT_TRACK_ADDED", "TRAJECTORY_DIR_ADDED",
                          "MARKER_CHANGE", "DRAW_EVERYTHING"]:

                # Convert lat/lon data into map coordinates, plot the item path
                # and set the visibility flag. A handle on the plot is stored
                # via the setplotInstance() method, this allows to later switch
                # on/off the visibility.
                x, y = self.__call__(item.getLonVariable().getVariableData(),
                                     item.getLatVariable().getVariableData())

                if mode != "MARKER_CHANGE":
                    # Remove old plot instances.
                    try:
                        instances = item.getGxElementProperty("lineplot",
                                                              u"instance::{}".format(self.identifier))
                        for instance in instances:
                            instance.remove()
                    except (KeyError, ValueError) as error:
                        logging.debug("KeyError, ValueError Exception %s", error)
                        pass
                    # Plot new instances.
                    plot_instance = self.plot(x, y,
                                              color=item_properties["colour"],
                                              visible=item_properties["visible"],
                                              linestyle=item_properties["linestyle"],
                                              linewidth=item_properties["linewidth"])
                    item.setGxElementProperty("lineplot",
                                              u"instance::{}".format(self.identifier),
                                              plot_instance)

                # Get the indexes for the time markers. If no time markers should
                # be drawn, 'None' is returned.
                try:
                    # Try to remove an existing time marker instance from
                    # the plot.
                    old_scatter_instance = item.getGxElementProperty("timemarker",
                                                                     u"instance::{}".format(self.identifier))
                    plt.setp(old_scatter_instance, visible=False)
                    old_scatter_instance.remove()
                except (KeyError, ValueError) as error:
                    logging.debug("KeyError, ValueError Exception %s", error)
                    pass
                imarker = item.getTimeMarkerIndexes()
                if imarker is not None:
                    scatter_instance = self.scatter(x[imarker], y[imarker],
                                                    s=20, c=item_properties["colour"],
                                                    visible=item_properties["visible"], linewidth=0)
                    item.setGxElementProperty("timemarker",
                                              u"instance::{}".format(self.identifier),
                                              scatter_instance)

        # Update the figure canvas.
        self.ax.figure.canvas.draw()

    def gcpoints2(self, lon1, lat1, lon2, lat2, del_s=100., map_coords=True):
        """
        The same as basemap.gcpoints(), but takes a distance interval del_s
        to space the points instead of a number of points.
        """
        # use great circle formula for a perfect sphere.
        gc = pyproj.Geod(a=self.rmajor, b=self.rminor)
        az12, az21, dist = gc.inv(lon1, lat1, lon2, lat2)
        npoints = int((dist + 0.5 * 1000. * del_s) / (1000. * del_s))
        lonlats = gc.npts(lon1, lat1, lon2, lat2, npoints)
        lons = [lon1]
        lats = [lat1]
        for lon, lat in lonlats:
            lons.append(lon)
            lats.append(lat)
        lons.append(lon2)
        lats.append(lat2)
        if map_coords:
            x, y = self(lons, lats)
        else:
            x, y = (lons, lats)
        return x, y

    def gcpoints_path(self, lons, lats, del_s=100., map_coords=True):
        """Same as gcpoints2, but for an entire path, i.e. multiple
           line segments. lons and lats are lists of waypoint coordinates.
        """
        # use great circle formula for a perfect sphere.
        gc = pyproj.Geod(a=self.rmajor, b=self.rminor)
        assert len(lons) == len(lats)
        assert len(lons) > 1
        gclons = [lons[0]]
        gclats = [lats[0]]
        for i in range(len(lons) - 1):
            az12, az21, dist = gc.inv(lons[i], lats[i], lons[i + 1], lats[i + 1])
            npoints = int((dist + 0.5 * 1000. * del_s) / (1000. * del_s))
            # BUG -- weird path in cyl projection on waypoint move
            # On some system configurations, the path is wrongly plotted when one
            # of the waypoints is moved by the user and the current projection is cylindric.
            # The weird thing is that when comparing cyl projection and stereo projection
            # (which works), the exact same arguments are passed to gc.npts(). Also, the
            # gc is initialised with the exact same a and b. Nevertheless, for the cyl
            # projection, gc.npts() returns lons that connect lon1 and lat2, not lon1 and
            # lon2 ... I cannot figure out why, maybe this is an issue in certain versions
            # of pyproj?? (mr, 16Oct2012)
            lonlats = gc.npts(lons[i], lats[i], lons[i + 1], lats[i + 1], npoints)
            # The cylindrical projection of matplotlib is not periodic, that means that
            # -170 longitude and 190 longitude are not identical. The gc projection however
            # assumes identity and maps all longitudes to -180 to 180. This is no issue for
            # most other projections.
            # The clean solution would be to have a periodic display, where the locations
            # and path are plotted periodically every 360 degree. As long as this is not
            # supported by matplotlib.basemap, we "hack" this to map the path to the
            # longitude range defined by the locations. This breaks potentially down in case
            # that the locations are too far apart (>180 degree), but this is not the typical
            # use case and will thus hopefully not pose a problem.
            if self.projection == "cyl" and npoints > 0:
                lonlats = np.asarray(lonlats)
                milon = min(lons[i], lons[i + 1])
                malon = max(lons[i], lons[i + 1])
                sel = lonlats[:, 0] < milon
                lonlats[sel, 0] += 360
                sel = lonlats[:, 0] > malon
                lonlats[sel, 0] -= 360

            for lon, lat in lonlats:
                gclons.append(lon)
                gclats.append(lat)
            gclons.append(lons[i + 1])
            gclats.append(lats[i + 1])
        if map_coords:
            x, y = self(gclons, gclats)
        else:
            x, y = (gclons, gclats)
        return x, y

    def drawgreatcircle_path(self, lons, lats, del_s=100., **kwargs):
        """
        """
        x, y = self.gcpoints_path(lons, lats, del_s=del_s)
        return self.plot(x, y, **kwargs)


class SatelliteOverpassPatch(object):
    """Represents a satellite overpass on the top view map (satellite
       track and, if available, swath).
    """

    # TODO: Derive this class from some Matplotlib actor class? Or create
    #       a new abstract base class for objects that can be drawn on the
    #       map -- they all should provide methods remove(), update(),
    #       etc. update() should automatically remap the object to new map
    #       coordinates.
    def __init__(self, mapcanvas, segment):
        """
        """
        self.map = mapcanvas
        self.segment = segment
        # Filter those time elements that correspond to masked positions -- this
        # way the indexes in self.utc correspond to those in self.sat.
        # np.ma.getmaskarray is necessary as ..mask only returns a scalar
        # "False" if the array contains no masked entries.
        self.utc = segment["utc"][~np.ma.getmaskarray(segment["satpos"])[:, 0]]
        self.sat = np.ma.compress_rows(segment["satpos"])
        self.sw_l = np.ma.compress_rows(segment["swath_left"])
        self.sw_r = np.ma.compress_rows(segment["swath_right"])
        self.trackline = None
        self.patch = None
        self.texts = []
        self.draw()

    def draw(self):
        """Do the actual plotting of the patch.
        """
        # Plot satellite track.
        sat = np.copy(self.sat)
        sat[:, 0], sat[:, 1] = self.map(sat[:, 0], sat[:, 1])
        self.trackline = self.map.plot(sat[:, 0], sat[:, 1], zorder=10,
                                       marker='+', markerfacecolor='g')

        # Plot polygon patch that represents the swath of the sensor.
        sw_l = self.sw_l
        sw_r = self.sw_r
        Path = mpath.Path
        pathdata = [(Path.MOVETO, self.map(sw_l[0, 0], sw_l[0, 1]))]
        for point in sw_l[1:]:
            pathdata.append((Path.LINETO, self.map(point[0], point[1])))
        for point in sw_r[::-1]:
            pathdata.append((Path.LINETO, self.map(point[0], point[1])))
        codes, verts = list(zip(*pathdata))
        path = mpl_pi.PathH(verts, codes, map=self.map)
        patch = mpatches.PathPatch(path, facecolor='yellow',
                                   edgecolor='yellow', alpha=0.4, zorder=10)
        self.patch = patch
        self.map.ax.add_patch(patch)

        # Draw text labels.
        self.texts.append(self.map.ax.text(sat[0, 0], sat[0, 1],
                                           self.utc[0].strftime("%H:%M:%S"),
                                           zorder=10,
                                           bbox=dict(facecolor='white',
                                                     alpha=0.5,
                                                     edgecolor='none')))
        self.texts.append(self.map.ax.text(sat[-1, 0], sat[-1, 1],
                                           self.utc[-1].strftime("%H:%M:%S"),
                                           zorder=10,
                                           bbox=dict(facecolor='white',
                                                     alpha=0.5,
                                                     edgecolor='none')))

        self.map.ax.figure.canvas.draw()

    def update(self):
        """Removes the current plot of the patch and redraws the patch.
           This is necessary, for instance, when the map projection and/or
           extent has been changed.
        """
        self.remove()
        self.draw()

    def remove(self):
        """Remove this satellite patch from the map canvas.
        """
        if self.trackline is not None:
            for element in self.trackline:
                element.remove()
            self.trackline = None
        if self.patch is not None:
            self.patch.remove()
            self.patch = None
        for element in self.texts:
            # Removal of text elements sometimes fails. I don't know why,
            # the plots look fine nevertheless.
            try:
                element.remove()
            except Exception as ex:
                logging.error("Wildcard exception caught: %s %s", type(ex), ex)


class KMLPatch(object):
    """
    Represents a KML overlay.

    KML overlay implementation is currently very crude and basic and most features are not supported.
    """

    def __init__(self, mapcanvas, kml, overwrite=False, color="red", linewidth=1):
        self.map = mapcanvas
        self.kml = kml
        self.patches = []
        self.color = color
        self.linewidth = linewidth
        self.overwrite = overwrite
        self.draw()

    def compute_xy(self, coordinates):
        coords = str(coordinates).split()
        lons, lats = [[float(_x.split(",")[_i]) for _x in coords] for _i in range(2)]
        return self.map(lons, lats)

    def add_polygon(self, polygon, style, _):
        """
        Plot KML polygons

        :param polygon: pykml object specifying a polygon
        """
        kwargs = self.styles.get(style, {}).get("PolyStyle", {"linewidth": self.linewidth, "color": self.color})
        for boundary in ["outerBoundaryIs", "innerBoundaryIs"]:
            if hasattr(polygon, boundary):
                x, y = self.compute_xy(getattr(polygon, boundary).LinearRing.coordinates)
                self.patches.append(self.map.plot(x, y, "-", zorder=10, **kwargs))

    def add_point(self, point, style, name):
        """
        Plot KML point

        :param point: pykml object specifying point
        :param name: name of placemark for annotation
        """
        x, y = self.compute_xy(point.coordinates)
        self.patches.append(self.map.plot(x[0], y[0], "o", zorder=10, color=self.color))
        if name is not None:
            self.patches.append([self.map.ax.annotate(name, xy=(x[0], y[0]), xycoords="data",
                                xytext=(5, 5), textcoords='offset points', zorder=10,
                                path_effects=[patheffects.withStroke(linewidth=2, foreground='w')])])

    def add_line(self, line, style, _):
        """
        Plot KML line

        :param line: pykml LineString object
        """
        kwargs = self.styles.get(style, {}).get("LineStyle", {"linewidth": self.linewidth, "color": self.color})
        x, y = self.compute_xy(line.coordinates)
        self.patches.append(self.map.plot(x, y, "-", zorder=10, **kwargs))

    def parse_geometries(self, placemark, style=None, name=None):
        if style.startswith("#"):
            # Remove # at beginning of style marking a locally defined style.
            # general urls for styles are not supported
            style = style[1:]
        for attr_name, method in (
                ("Point", self.add_point),
                ("Polygon", self.add_polygon),
                ("LineString", self.add_line),
                ("MultiGeometry", self.parse_geometries)):
            for attr in getattr(placemark, attr_name, []):
                logging.debug("Found %s", attr_name)
                method(attr, style, name)

    def parse_placemarks(self, level):
        for placemark in getattr(level, "Placemark", []):
            name = getattr(placemark, "name", None)
            style = str(getattr(placemark, "styleUrl", ""))
            logging.debug("Placemark: %s %s", style, name)
            self.parse_geometries(placemark, style, name)
        for folder in getattr(level, "Folder", []):
            name = getattr(folder, "name", None)
            logging.debug("Folder: %s", name)
            self.parse_placemarks(folder)

    def get_style_params(self, style):
        result = {
            "color": str(getattr(style, "color", "")),
            "linewidth": float(getattr(style, "width", self.linewidth))
        }
        if not result["color"] or len(result["color"]) != 8:
            result["color"] = self.color
        else:
            result["color"] = [(int(result["color"][i:i + 2], 16) / 255.) for i in range(0, 8, 2)]
        return result

    def parse_styles(self, level):
        for style in getattr(level, "Style", []):
            name = style.attrib.get("id")
            if name is None:
                continue
            self.styles[name] = {
                "LineStyle": self.get_style_params(getattr(style, "LineStyle", None)),
                "PolyStyle": self.get_style_params(getattr(style, "PolyStyle", None))
            }
        for folder in getattr(level, "Folder", []):
            name = getattr(folder, "name", None)
            logging.debug("Folder: %s", name)
            self.parse_styles(folder)

    def draw(self):
        """Do the actual plotting of the patch.
        """
        # Plot satellite track.
        self.styles = {}
        if not self.overwrite:
            self.parse_styles(self.kml.Document)
        self.parse_placemarks(self.kml.Document)

        self.map.ax.figure.canvas.draw()

    def update(self, overwrite=None, color=None, linewidth=None):
        """Removes the current plot of the patch and redraws the patch.
           This is necessary, for instance, when the map projection and/or
           extent has been changed.
        """
        if overwrite is not None:
            self.overwrite = overwrite
        if color is not None:
            self.color = color
        if linewidth is not None:
            self.linewidth = linewidth
        self.remove()
        self.draw()

    def remove(self):
        """Remove this satellite patch from the map canvas.
        """
        for patch in self.patches:
            for element in patch:
                element.remove()
        self.patches = []
        self.map.ax.figure.canvas.draw()



# -*- coding: utf-8 -*-
"""

    mslib.msui._tests.test_trajectory_tool
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to tests msui.trajectories_tool

    This file is part of mss.

    :copyright: Copyright 2017 Joern Ungermann
    :copyright: Copyright 2017-2018 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
import sys
import mock
import pytest
from mslib.msui.mss_qt import QtWidgets, QtTest, QtCore
import mslib.msui.trajectories_tool as tt
import mslib.msui.mss_pyui as mss_pyui
from mslib.msui.trajectory_item_tree import HAVE_NAPPY


class Test_TrajectoriesTool(object):
    sample_path = os.path.join(os.path.dirname(__file__), "..", "..", "..", "docs", "samples")

    def setup(self):
        self.application = QtWidgets.QApplication(sys.argv)
        self.listViews = mock.Mock()
        self.listTools = mock.Mock()
        self.listViews.findItems = mock.Mock(side_effect=lambda x, y: [])
        self.listTools.findItems = mock.Mock(side_effect=lambda x, y: [])
        self.window = tt.MSSTrajectoriesToolWindow(listviews=self.listViews, listtools=self.listTools)
        self.window.show()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWaitForWindowExposed(self.window)
        QtWidgets.QApplication.processEvents()

    def teardown(self):
        self.window.hide()
        QtWidgets.QApplication.processEvents()
        self.application.quit()
        QtWidgets.QApplication.processEvents()

    @mock.patch("mslib.msui.mss_qt.QtWidgets.QMessageBox.critical")
    def test_show(self, mockcrit):
        assert mockcrit.call_count == 0

    @mock.patch("mslib.msui.trajectories_tool.get_existing_directory",
                return_value=os.path.join(sample_path, "trajectories"))
    def test_load_trajectories(self, mockopen):
        if sys.version_info.major > 2:
            pytest.skip("can't load example pickle file")
        pytest.skip('{RuntimeError} dictionary changed size during iteration')
        self.window.actionOpenTrajectories.trigger()
        QtWidgets.QApplication.processEvents()
        assert mockopen.call_count == 1

    @mock.patch("mslib.msui.trajectories_tool.get_open_filename",
                return_value=os.path.join(sample_path, "nas", "sample.nas"))
    def test_load_nas(self, mockopen):
        if not HAVE_NAPPY:
            pytest.skip("nappy not available")
        self.window.actionOpenFlightTrack.trigger()
        QtWidgets.QApplication.processEvents()
        assert mockopen.call_count == 1

    @mock.patch("mslib.msui.trajectories_tool.get_open_filename",
                return_value=os.path.join(sample_path, "kml", "line.kml"))
    def test_load_nas_kml(self, mockopen):
        self.window.actionOpenFlightTrack.trigger()
        QtWidgets.QApplication.processEvents()
        assert mockopen.call_count == 1


class Test_TrajectoryToolComples(object):
    sample_path = os.path.join(os.path.dirname(__file__), "..", "..", "..", "docs", "samples")

    def setup(self):
        self.application = QtWidgets.QApplication(sys.argv)
        self.window = mss_pyui.MSSMainWindow()
        self.window.create_new_flight_track()
        self.window.show()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWaitForWindowExposed(self.window)
        QtWidgets.QApplication.processEvents()
        self.window.actionTopView.trigger()
        QtWidgets.QApplication.processEvents()
        self.window.actionTrajectoryToolLagranto.trigger()
        QtWidgets.QApplication.processEvents()
        self.topview = self.window.listViews.item(0).window
        self.trajtool = self.window.listTools.item(0).window

    def teardown(self):
        for i in range(self.window.listViews.count()):
            self.window.listViews.item(i).window.hide()
        for i in range(self.window.listTools.count()):
            self.window.listTools.item(i).window.hide()
        self.window.hide()
        QtWidgets.QApplication.processEvents()
        self.application.quit()
        QtWidgets.QApplication.processEvents()

    @mock.patch("mslib.msui.trajectories_tool.get_open_filename",
                return_value=os.path.join(sample_path, "nas", "sample.nas"))
    def test_show_nas(self, mockopen):
        if not HAVE_NAPPY:
            pytest.skip("nappy not available")
        self.trajtool.actionOpenFlightTrack.trigger()
        QtWidgets.QApplication.processEvents()
        assert mockopen.call_count == 1
        self.trajtool.cbPlotInView.setCurrentIndex(1)
        QtWidgets.QApplication.processEvents()
        root = self.trajtool.traj_item_tree.getRootItem()
        flighttrack = self.trajtool.tvVisibleElements.visualRect(
            self.trajtool.traj_item_tree.createIndex(0, 0, root.childItems[0].childItems[0]))
        QtTest.QTest.mouseClick(
            self.trajtool.tvVisibleElements.viewport(),
            QtCore.Qt.LeftButton, QtCore.Qt.NoModifier, flighttrack.center())
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.mouseClick(self.trajtool.btPlotInView, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()

    @mock.patch("mslib.msui.trajectories_tool.get_existing_directory",
                return_value=os.path.join(sample_path, "trajectories"))
    def test_show_trajectories(self, mockopen):
        if sys.version_info.major > 2:
            pytest.skip("can't load example pickle file")
        pytest.skip('{RuntimeError} dictionary changed size during iteration')
        self.trajtool.actionOpenTrajectories.trigger()
        QtWidgets.QApplication.processEvents()
        assert mockopen.call_count == 1
        self.trajtool.cbPlotInView.setCurrentIndex(1)
        QtWidgets.QApplication.processEvents()
        root = self.trajtool.traj_item_tree.getRootItem()
        self.trajtool.tvVisibleElements.expandAll()
        flighttrack = self.trajtool.tvVisibleElements.visualRect(
            self.trajtool.traj_item_tree.createIndex(0, 0, root.childItems[1].childItems[0].childItems[0]))
        QtTest.QTest.mouseClick(
            self.trajtool.tvVisibleElements.viewport(),
            QtCore.Qt.LeftButton, QtCore.Qt.NoModifier, flighttrack.center())
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.mouseClick(self.trajtool.btPlotInView, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
